/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2022 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       MiddleWare plug-in                                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       Script version: V1.0.0 Rev. 1                                *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Script for middleware windows for SEGGER Ozone
*/

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/
var _PagePathSeparator  = "/";
var _NumPages           = 12;
var _CurrentRowOnPage   = [];

var _Cache_FS_Global                               = {};
var _Cache_FS_Global_IsValid                       = false;
var _Cache_FS_Global_SectorBuffer                  = [];
var _Cache_FS_Global_SectorBuffer_IsValid          = false;
var _Cache_FS_Global_DirEntrySectorBuffer          = [];
var _Cache_FS_Global_DirEntrySectorBuffer_IsValid  = false;
var _Cache_FS_Global_Copyright                     = {};
var _Cache_FS_Global_Copyright_IsValid             = false;
var _Cache_FileObject                              = {};
var _Cache_FileObject_IsValid                      = false;
var _Cache_VolumeInfos                             = [];
var _Cache_VolumeInfos_IsValid                     = false;
var _Cache_FileHandleInfos                         = [];
var _Cache_FileHandleInfos_IsValid                 = false;
var _Cache_FileObjectInfos                         = [];
var _Cache_FileObjectInfos_IsValid                 = false;
var _Cache_SectorBufferObject                      = {};
var _Cache_SectorBufferObject_IsValid              = false;
var _Cache_FAT_pDirEntryAPI                        = {};
var _Cache_FAT_pDirEntryAPI_IsValid                = false;
var _Cache_FAT_LFN_API_Adr                         = 0;
var _Cache_FAT_LFN_API_Adr_IsValid                 = false;

var _Feature_EmFilePresent                         = false;
var _Feature_FS_SUPPORT_DEINIT                     = false;
var _Feature_FS_SUPPORT_FILE_BUFFER                = false;
var _Feature_FS_SUPPORT_SECTOR_BUFFER_CACHE        = false;
var _Feature_FS_SUPPORT_PROFILE                    = false;
var _Feature_FS_SUPPORT_FREE_SECTOR                = false;
var _Feature_FS_SUPPORT_JOURNAL                    = false;
var _Feature_FS_SUPPORT_JOURNAL_FILENAME           = false;
var _Feature_FS_SUPPORT_CACHE                      = false;
var _Feature_FS_SUPPORT_BUSY_LED                   = false;
var _Feature_FS_SUPPORT_CHECK_MEMORY               = false;
var _Feature_FS_STORAGE_SUPPORT_DEVICE_ACTIVITY    = false;
var _Feature_FS_SUPPORT_VOLUME_ALIAS               = "Disabled";
var _Feature_FS_SUPPORT_EXT_MEM_MANAGER            = false;
var _Feature_FS_MULTI_HANDLE_SAFE                  = false;
var _Feature_FS_EFS_SUPPORT_STATUS_SECTOR          = false;
var _Feature_FS_EFS_SUPPORT_FREE_CLUSTER_CACHE     = false;
var _Feature_FS_FAT_USE_FSINFO_SECTOR              = false;
var _Feature_FS_FAT_UPDATE_DIRTY_FLAG              = false;
var _Feature_FS_FAT_SUPPORT_FREE_CLUSTER_CACHE     = false;
var _Feature_FS_EFS_SUPPORT_DIRENTRY_BUFFERS       = false;
var _Feature_FS_SUPPORT_ENCRYPTION                 = false;
var _Feature_FS_OS_LOCKING                         = "FS_OS_LOCKING_NONE";
var _Feature_FS_SUPPORT_FAT                        = false;
var _Feature_FS_SUPPORT_EFS                        = false;
var _Feature_FS_SUPPORT_MULTIPLE_FS                = false;
var _Feature_FS_VERIFY_WRITE                       = false;
var _Feature_FS_SUPPRESS_EOF_ERROR                 = false;
var _Feature_FS_SUPPORT_POSIX                      = false;
var _Feature_FS_SUPPORT_MBCS                       = false;
var _Feature_FS_SUPPORT_EXT_ASCII                  = false;
var _Feature_FS_SUPPORT_FILE_NAME_ENCODING         = false;
var _Feature_FS_MAINTAIN_FAT_COPY                  = false;
var _Feature_FS_JOURNAL_ENABLE_STATS               = false;
var _Feature_FS_JOURNAL_SUPPORT_FREE_SECTOR        = false;
var _Feature_FS_JOURNAL_SUPPORT_FAST_SECTOR_SEARCH = false;
var _Feature_FS_JOURNAL_OPTIMIZE_SPACE_USAGE       = false;
var _Feature_FS_FAT_OPTIMIZE_DELETE                = false;
var _Feature_FS_FAT_OPTIMIZE_LINEAR_ACCESS         = false;
var _Feature_FS_FAT_PERMIT_RO_FILE_MOVE            = false;
var _Feature_FS_FAT_LFN_LOWER_CASE_SHORT_NAMES     = false;
var _Feature_FS_FAT_CHECK_UNUSED_DIR_ENTRIES       = false;
var _Feature_FS_EFS_OPTIMIZE_DELETE                = false;
var _Feature_FS_EFS_CASE_SENSITIVE                 = false;
var _Feature_FS_EFS_OPTIMIZE_LINEAR_ACCESS         = false;
var _Feature_FS_STORAGE_ENABLE_STAT_COUNTERS       = false;
var _Feature_FS_STORAGE_SUPPORT_DEVICE_ACTIVITY    = false;
var _Feature_FS_OS_SUPPORT_RUNTIME_CONFIG          = false;
var _Feature_FS_NAND_ENABLE_STATS                  = false;
var _Feature_FS_FAT_SUPPORT_LFN                    = false;
var _Feature_LongFilenameSupport                   = "";

var _MiddleWareNotFoundMsgPrinted                  = false;

/*********************************************************************
*
*       Common ToStr functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _ByteArrayToStr()
*
*  Function description:
*    Translates a byte array into a string.
*/
function _ByteArrayToStr(Value) {
  var Result;

  Result = "";
  for(var i = 0; i < Value.length; i++) {
    if(Value[i] != 0){
      Result += String.fromCharCode(Value[i]);
    } else {
      break;
    }
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrBoolYesNo()
*
*  Function description:
*    Translates value to "Yes" or "No"
*/
function _ToStrBoolYesNo(Value){
  var Result;

  if(Value == 0){
    Result = "No";
  } else {
    Result = "Yes";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrBoolEnabledDisable()
*
*  Function description:
*    Translates Value to "Enabled" or "Disabled"
*/
function _ToStrBoolEnabledDisable(Value){
  var Result;

  if(Value == 0){
    Result = "Disabled";
  } else {
    Result = "Enabled";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrBoolTrueFalse()
*
*  Function description:
*    Translates Value to "True" or "False"
*/
function _ToStrBoolTrueFalse(Value){
  var Result;

  if(Value == 0){
    Result = "False";
  } else {
    Result = "True";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrBoolSetClear()
*
*  Function description:
*    Translates Value to "Set" or "Clear"
*/
function _ToStrBoolSetClear(Value){
  var Result;

  if(Value == 0){
    Result = "Clear";
  } else {
    Result = "Set";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStr()
*
*  Function description:
*    Translates Value to string
*/
function _ToStr(Value) {
  var Result;

  Result = Value.toString();
  return Result;
}

/*********************************************************************
*
*       _ToStrSectorSuffix()
*
*  Function description:
*    Translates value to string
*/
function _ToStrSectorSuffix (Value) {
  var Result;

  Result = Value.toString();
  if (Value == 1) {
    Result += " sector";
  } else {
    Result += " sectors";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrSectorPrefix()
*
*  Function description:
*    Translates value to string
*/
function _ToStrSectorPrefix (Value) {
  var Result;

  Result = "sector " + Value;
  return Result;
}

/*********************************************************************
*
*       _ToStrClusterSuffix()
*
*  Function description:
*    Translates value to string
*/
function _ToStrClusterSuffix(Value) {
  var Result;

  Result = Value.toString();
  if (Value == 1) {
    Result += " cluster";
  } else {
    Result += " clusters";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrClusterPrefix()
*
*  Function description:
*    Translates value to string
*/
function _ToStrClusterPrefix(Value) {
  var Result;

  Result = "cluster " + Value;
  return Result;
}

/*********************************************************************
*
*       _ToStrEntrySuffix()
*
*  Function description:
*    Translates value to string
*/
function _ToStrEntrySuffix(Value) {
  var Result;

  Result = Value.toString();
  if (Value == 1) {
    Result += " entry";
  } else {
    Result += " entries";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrByteSuffix()
*
*  Function description:
*    Translates Value to string
*/
function _ToStrByteSuffix(Value) {
  var Result;

  Result = Value.toString();
  if (Value == 1) {
    Result += " byte";
  } else {
    Result += " bytes";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrKMGByteSuffix()
*
*  Function description:
*    Translates Value to string
*/
function _ToStrKMGByteSuffix(Value) {
  var Result;

  NewValue = Value;
  if (NewValue < 1024) {
    Result = _ToStrByteSuffix(NewValue);
  } else {
    NewValue = Value / 1024;
    if (NewValue < 1024) {
      Result = NewValue.toFixed(2).toString() + " Kbytes";
    } else {
      NewValue = NewValue / 1024;
      if (NewValue < 1024) {
        Result = NewValue.toFixed(2).toString() + " Mbytes";
      } else {
        NewValue = NewValue / 1024;
        if (NewValue < 1024) {
          Result = NewValue.toFixed(2).toString() + " Gbytes";
        } else {
          NewValue = NewValue / 1024;
          Result   = NewValue.toFixed(2).toString() + " Tbytes";
        }
      }
    }
  }
  return Result;
}

/*********************************************************************
*
*       _PowerOf2
*
*  Function description:
*    Calculates 2 to the power of x, whereas x is the value passed as
*    parameter.
*    Value must be an integer in the range [0..[.
*/
function _PowerOf2(Value) {
  var Cnt;
  var Result;

  Result = 1;
  for (Cnt = 1; Cnt <= Value; Cnt++) {
    Result = Result * 2;
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrPowerOf2ByteSuffix()
*
*  Function description:
*    Translates Value to string
*/
function _ToStrPowerOf2ByteSuffix(Value) {
  var NumBytes;
  var Result;

  NumBytes = _PowerOf2(Value);
  Result = _ToStrByteSuffix(NumBytes);
  return Result;
}

/*********************************************************************
*
*       _ToStrHex()
*
*  Function description:
*    Translates Value to hex-string
*/
function _ToStrHex(Value) {
  var Result;

  Result = "0x" + Value.toString(16);
  return Result;
}

/*********************************************************************
*
*       _ToStrOrEmptyTag()
*
*  Function description:
*    Translates string to string or "<empty>"-tag, if string is empty
*/
function _ToStrOrEmptyTag(Value) {
  var Result;

  Result = Value;
  if (Result == "") {
    Result = "<empty>";
  }
  return Result;
}

/*********************************************************************
*
*       _ByteArrayToStrOrEmptyTag()
*
*  Function description:
*    Translates string to string or "<empty>"-tag, if string is empty
*/
function _ByteArrayToStrOrEmptyTag(Value) {
  var Result;

  Result = _ByteArrayToStr(Value);
  if (Result == "") {
    Result = "<empty>";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrFuncNameOrUnknownTag()
*
*  Function description:
*    Translates string to string or "<empty>"-tag, if string is empty
*/
function _ToStrFuncNameOrUnknownTag(Value) {
  var Result;
  var FuncName;

  FuncName = "";
  if (typeof Value == "number") {
    FuncName = Debug.getSymbol(Value);
  }
  if ((Value == 0) ||
      (FuncName == "") ||
      (FuncName == "?") ||
      (FuncName == "undefined")){
    Result = "<empty>";
  } else {
    Result = FuncName + "()";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrIdentifierOrUnknownTag()
*
*  Functions description:
*    Translates value to string or "<unknwon>"-tag, if Value is a
*    number instead of a string.
*/
function _ToStrIdentifierOrUnknownTag(Value) {
  var Result;

  if(typeof Value == "number") {
    Result = "<unknown> (" + Value + ")";
  } else {
    Result = Value.toString();
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrFileBufferFlags()
*
*  Functions description:
*    Extracts the "valid" flag from file buffer flags.
*/
function _ToStrFileBufferFlags(Value) {
  var Result;

  if (Value & 0x01) {
    Result ="Read/write"
  } else {
    Result = "Read"
  }
  if (Value & 0x02) {
    Result += ", aligned"
  }
  return Result;
}

/*********************************************************************
*
*       _GetVolumeName()
*
*  Functions description:
*    Get the name of a volume by its Id (address of volume info).
*
*  Additional information
*    Volume info cache needs to be up to date.
*    Call _UpdateTargetInfoVolume() before use!
*/
function _GetVolumeName(Addr)  {
  var VolumeInfoRecord;
  var Alias;
  var sValue;

  for (var Idx = 0; Idx < _Cache_VolumeInfos.length; Idx++) {
    VolumeInfoRecord = _Cache_VolumeInfos[Idx];
    if (VolumeInfoRecord["Addr"] == Addr) {
      Alias = VolumeInfoRecord["Alias"];
      if (VolumeInfoRecord["Alias"] != "") {
        sValue = Alias;
        return sValue;
      } else {
        sValue = "Volume " + Idx;
        return sValue;
      }
    }
  }
  return undefined;
}

/*********************************************************************
*
*       _GetFileNameFromFileObject()
*
*  Functions description:
*    Get the name of a volume by its Id (address of volume info).
*
*  Additional information
*    Volume info cache needs to be up to date.
*    Call _UpdateTargetInfoVolume() before use!
*/
function _GetFileNameFromFileObject(CurrentFileObject)  {
  var Result;
  var Value;

  Result = "";
  if(_Feature_FS_MULTI_HANDLE_SAFE == true) {
    Value  = CurrentFileObject["acFullFileName"];
    Result = _ByteArrayToStr(Value);
  }
  return Result;
}

/*********************************************************************
*
*       _GetFileNameFromFileObjectAddr()
*
*  Functions description:
*    Get the name of a volume by its Id (address of volume info).
*
*  Additional information
*    Volume info cache needs to be up to date.
*    Call _UpdateTargetInfoVolume() before use!
*/
function _GetFileNameFromFileObjectAddr(Addr)  {
  var Result;
  var FileObjectInfoRecord;
  var CurrentFileObject;
  var sValue;

  Result = _ToStrHex(Addr);
  if(_Feature_FS_MULTI_HANDLE_SAFE == true) {
    for (var Idx = 0; Idx < _Cache_FileObjectInfos.length; Idx++) {
      FileObjectInfoRecord = _Cache_FileObjectInfos[Idx];
      if (FileObjectInfoRecord["Addr"] == Addr) {
        CurrentFileObject = FileObjectInfoRecord["FileObject"];
        sValue            = _GetFileNameFromFileObject(CurrentFileObject);
        if(sValue != "") {
          Result = sValue;
        }
      }
    }
  }
  return Result;
}

/*********************************************************************
*
*       Common GetRow functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _GetFeatureNotEnabledStr()
*
*  Function description:
*    Creates a string containing "<n/a>" tag
*/
function _GetFeatureNotEnabledStr() {
  var Result;
  
  Result = "<n/a>";
  return Result;
}

/*********************************************************************
*
*       _GetFeatureNotEnabledVRow()
*
*  Function description:
*    Creates a row of entries containing "<feature not enabled>" tag
*/
function _GetFeatureNotEnabledVRow(ItemName, NumCols) {
  var Result;
  var sValue;

  Result = [];
  Result.push(ItemName);
  if (_Feature_EmFilePresent == true) {
    sValue = _GetFeatureNotEnabledStr();
    for (var Idx = 0; Idx < NumCols; Idx++) {
      Result.push(sValue);
    }
  }
  return Result;
}

/*********************************************************************
*
*       _GetFeatureNotEnabledHRow()
*
*  Function description:
*    Creates a row of entries containing "<n/a>" tag
*/
function _GetFeatureNotEnabledHRow(NumCols) {
  var Result;
  var sValue;
  
  Result = [];
  if (_Feature_EmFilePresent == true) {
    sValue = _GetFeatureNotEnabledStr;
    for (var Idx = 0; Idx < NumCols; Idx++) {
      Result.push(sValue);
    }
  }
  return Result;
}

/*********************************************************************
*
*       _GetHeadlineRow()
*
*  Function description:
*    Creates a row of entries containing the item name followed by 
*    empty cells.
*/
function _GetHeadlineRow(ItemName, NumCols) {
  var Result;

  Result = [];
  Result.push(ItemName);
  for (var Idx = 0; Idx < NumCols; Idx++) {
    Result.push("");
  }
  return Result;
}

/*********************************************************************
*
*       Common bit-field processing functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _GetBitValue()
*
*  Function description:
*    Extracts a bitmask from a Value and returns the value of the
*    extracted bits
*/
function _GetBitValue(Value, BitMask, BitOffset){
  var Result;
  var BitValue;

  BitValue = (Value & BitMask) >> BitOffset;
  Result   = _ToStrBoolSetClear(BitValue);
  return Result;
}

/*********************************************************************
*
*       Common target information retrieval functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _InvalidateCaches()
*
*  Functions description:
*    Marks all local caches as invalid, thus allowing update upon next
*    occasion.
*/
function _InvalidateCaches() {
  _Cache_FS_Global_IsValid              = false;
  _Cache_FS_Global_SectorBuffer_IsValid = false;
  _Cache_FS_Global_Copyright_IsValid    = false;
  _Cache_FileObject_IsValid             = false;
  _Cache_VolumeInfos_IsValid            = false;
  _Cache_FileHandleInfos_IsValid        = false;
  _Cache_FileObjectInfos_IsValid        = false;
  _Cache_SectorBufferObject_IsValid     = false;
}

/*********************************************************************
*
*       _UpdateTargetInfoFsGlobal()
*
*  Functions description:
*    Retrieves the global FS info record from the target
*/
function _UpdateTargetInfoFsGlobal() {
  if (_Cache_FS_Global_IsValid == false) {
    _Cache_FS_Global          = Debug.evaluate("FS_Global");
    _Cache_FS_Global_IsValid  = true;
  }
}

/*********************************************************************
*
*       _UpdateTargetInfoFsGlobalCopyright()
*
*  Functions description:
*    Retrieves the global FS info record from the target
*/
function _UpdateTargetInfoFsGlobalCopyright() {
  if (_Feature_EmFilePresent == true) {
    _UpdateTargetInfoFsGlobal();
    if (_Cache_FS_Global_Copyright_IsValid == false) {
      _Cache_FS_Global_Copyright = Debug.evaluate("FS_Global.sCopyright");
      _Cache_FS_Global_Copyright_IsValid = true;
    }
  }
}

/*********************************************************************
*
*       _UpdateTargetInfoSectorBufferObject()
*
*  Functions description:
*    Retrieves a FileObject record from the target.
*    The file object can be used solely for checking for the presence
*    of members since it is always retrieved from address 0x0 which
*    in general does not house such a structure.
*/
function _UpdateTargetInfoSectorBufferObject() {
  if (_Cache_SectorBufferObject_IsValid == false) {
    _Cache_SectorBufferObject         = Debug.evaluate("*(SECTOR_BUFFER*)0x00000000");
    _Cache_SectorBufferObject_IsValid = true;
  }
}

/*********************************************************************
*
*       _UpdateTargetInfoFileObject()
*
*  Functions description:
*    Retrieves a FileObject record from the target.
*    The file object can be used solely for checking for the presence
*    of members since it is always retrieved from address 0x0 which
*    in general does not house such a structure.
*/
function _UpdateTargetInfoFileObject() {
  if (_Cache_FileObject_IsValid == false) {
    _Cache_FileObject         = Debug.evaluate("*(FS_FILE_OBJ*)0x00000000");
    _Cache_FileObject_IsValid = true;
  }
}

/*********************************************************************
*
*       _UpdateTargetInfoVolume()
*
*  Functions description:
*    Retrieves the volume info records from the target
*/
function _UpdateTargetInfoVolume() {
  var NumVolumes;
  var Expression;
  var NextVolumeAddr;
  var CurrentVolumeInfoRecord;
  var CurrentVolume;
  var CurrentFsApiAddr;
  var CurrentFsApiName;
  var CurrentFsApi;

  if (_Feature_EmFilePresent == true) {
    _UpdateTargetInfoFsGlobal();
    if (_Cache_VolumeInfos_IsValid == false) {
      _Cache_VolumeInfos  = [];
      NumVolumes = _Cache_FS_Global["NumVolumes"];
      if (NumVolumes > 0) {
        Expression = "&(FS_Global.FirstVolume)";
        NextVolumeAddr = Debug.evaluate(Expression);
        do {
          CurrentVolumeInfoRecord                    = {};
          CurrentVolumeInfoRecord["Addr"]            = 0x0;
          CurrentVolumeInfoRecord["Volume"]          = {};
          CurrentVolumeInfoRecord["HasFsApi"]        = false;
          CurrentVolumeInfoRecord["FsApiAddr"]       = 0x0;
          CurrentVolumeInfoRecord["FsApi"]           = {};
          CurrentVolumeInfoRecord["IsFAT"]           = false;
          CurrentVolumeInfoRecord["IsEFS"]           = false;
          if ((_Feature_FS_SUPPORT_FAT == true) &&
              (_Feature_FS_SUPPORT_EFS == false)) {
            CurrentVolumeInfoRecord["IsFAT"]         = true;
            CurrentVolumeInfoRecord["IsEFS"]         = false;
          }
          if ((_Feature_FS_SUPPORT_FAT == false) &&
              (_Feature_FS_SUPPORT_EFS == true)) {
            CurrentVolumeInfoRecord["IsFAT"]         = false;
            CurrentVolumeInfoRecord["IsEFS"]         = true;
          }
          if (NextVolumeAddr != 0) {
            var CurrentVolumeAlias;

            CurrentVolumeInfoRecord["Addr"]          = NextVolumeAddr;
            Expression                               = "*(FS_VOLUME*)" + _ToStrHex(NextVolumeAddr);
            CurrentVolume                            = Debug.evaluate(Expression);
            CurrentVolumeInfoRecord["Volume"]        = CurrentVolume;
            if (_Feature_FS_SUPPORT_VOLUME_ALIAS == true) {
              Expression                             = "(*(FS_VOLUME*)" + _ToStrHex(NextVolumeAddr) + ").acAlias";
              CurrentVolumeAlias                     = Debug.evaluate(Expression);
            } else if(_Feature_FS_SUPPORT_VOLUME_ALIAS == "Enabled, othersAliasName") {
              Expression                             = "(*(FS_VOLUME*)" + _ToStrHex(NextVolumeAddr) + ").sAlias";
              CurrentVolumeAlias                     = Debug.evaluate(Expression);
            } else {
              CurrentVolumeAlias                     = "";
            }
            CurrentVolumeInfoRecord["Alias"]         = CurrentVolumeAlias;
            if (_Feature_FS_SUPPORT_MULTIPLE_FS == true) {
              CurrentFsApiAddr                       = CurrentVolume["pFS_API"];
              if (CurrentFsApiAddr != 0) {
                CurrentVolumeInfoRecord["HasFsApi"]  = true;
                CurrentVolumeInfoRecord["FsApiAddr"] = CurrentFsApiAddr;
                CurrentFsApiName                     = Debug.getSymbol(CurrentFsApiAddr);
                Expression                           = "*(FS_FS_API *)" + _ToStrHex(CurrentFsApiAddr);
                CurrentFsApi                         = Debug.evaluate(Expression);
                CurrentVolumeInfoRecord["FsApi"]     = CurrentFsApi;
                if (CurrentFsApiName == "FS_FAT_API") {
                  CurrentVolumeInfoRecord["IsFAT"]   = true;
                  CurrentVolumeInfoRecord["IsEFS"]   = false;      
                } else if (CurrentFsApiName == "FS_EFS_API") {
                  CurrentVolumeInfoRecord["IsFAT"]   = false;
                  CurrentVolumeInfoRecord["IsEFS"]   = true;      
                } else {
                  CurrentVolumeInfoRecord["IsEFS"]   = false;      
                  CurrentVolumeInfoRecord["IsFAT"]   = false;      
                  TargetInterface.message("Unsupported FS-API found: Neither FAT nor EFS: " + CurrentFsApiName);
                }
              }
            }
            NextVolumeAddr                           = CurrentVolume["pNext"];
            _Cache_VolumeInfos.push(CurrentVolumeInfoRecord);
          }
        } while (NextVolumeAddr != 0);
      }
      _Cache_VolumeInfos_IsValid = true;
    }
  }
}

/*********************************************************************
*
*       _UpdateTargetInfoLfnSupportCheck()
*
*  Functions description:
*    Retrieves the information required for identifying of long file
*    name support being in place.
*/
function _UpdateTargetInfoLfnSupportCheck() {
  if (_Feature_EmFilePresent == true) {
    _Cache_FAT_pDirEntryAPI         = Debug.evaluate("FAT_pDirEntryAPI");
    _Cache_FAT_pDirEntryAPI_IsValid = true;
    _Cache_FAT_LFN_API_Adr          = Debug.evaluate("&FAT_LFN_API");
    _Cache_FAT_LFN_API_Adr_IsValid  = true;
  }
}

/*********************************************************************
*
*       Detection...
*
**********************************************************************
*/

/*********************************************************************
*
*       _MiddlewareDetection()
*
*  Function description:
*    Checks if the middleware is part of the current project. If not,
*    print a message onto the console just once.
*/
function _MiddlewareDetection() {
  _UpdateFeatureFlags();
  if (_Feature_EmFilePresent == false) {
    if (_MiddleWareNotFoundMsgPrinted == false) {
      TargetInterface.message("emFile plugin loaded but emFile middleware not part of current project (symbol FS_Global not found)");
      _MiddleWareNotFoundMsgPrinted = true;
    }
  }
}

/*********************************************************************
*
*       _LfnSupportDetection()
*
*  Function description:
*    Checks if the lonmg file name support is enabled and active.
*/
function _LfnSupportDetection() {
  _Feature_LongFilenameSupport = "Short file names only";
  _UpdateFeatureFlags();
  if (_Feature_FS_FAT_SUPPORT_LFN == true) {
    _UpdateTargetInfoLfnSupportCheck();
    if (_Cache_FAT_LFN_API_Adr != undefined) {
      if (_Cache_FAT_pDirEntryAPI == _Cache_FAT_LFN_API_Adr) {
        _Feature_LongFilenameSupport = "Long and short file names";
      }
    }
  }
}

/*********************************************************************
*
*       Feature flag detection
*
**********************************************************************
*/

/*********************************************************************
*
*       _UpdateFeatureFlags()
*
*  Function description:
*    Tests all known feature flags in target firmware for being enabled
*    or disabled
*/
function _UpdateFeatureFlags() {
  var FS_Global_pFirstOnExit;
  var FS_Global_FileBufferSize;
  var FS_Global_Profile;
  var FS_Global_NumDirEntryBuffers;
  var FS_Global_FirstVolume;
  var FS_Global_FirstVolume_pFS_API;
  var FS_Global_FirstVolume_FreeSector;
  var FS_Global_FirstVolume_Partition;
  var FS_Global_FirstVolume_Partition_Device;
  var FS_Global_FirstVolume_Partition_Device_Data;
  var FS_Global_FirstVolume_Partition_Device_Data_JournalData;
  var FS_Global_FirstVolume_Partition_Device_Data_pCacheData;
  var FS_Global_FirstVolume_Partition_Device_Data_pfSetBusyLED;
  var FS_Global_FirstVolume_Partition_Device_Data_pfMemCheck;
  var FS_Global_FirstVolume_Partition_Device_Data_pfOnDeviceActivity;
  var FS_Global_FirstVolume_acJournalFileName;
  var FS_Global_FirstVolume_acAlias;
  var FS_Global_FirstVolume_sAlias;
  var FS_Global_FirstVolume_FSInfo;
  var FS_Global_FirstVolume_FSInfo_EFSInfo;
  var FS_Global_FirstVolume_FSInfo_EFSInfo_IsNumFreeClustersInvalidated;
  var FS_Global_FirstVolume_FSInfo_EFSInfo_FreeClusterCache;
  var FS_Global_FirstVolume_FSInfo_FATInfo;
  var FS_Global_FirstVolume_FSInfo_FATInfo_FSInfoSector;
  var FS_Global_FirstVolume_FSInfo_FATInfo_IsDirty;
  var FS_Global_FirstVolume_FSInfo_FATInfo_FreeClusterCache;
  var FS_Global_MemManager;
  var FS_Global_MemManager_pfAlloc;
  var SectorBufferObject;
  var SectorBufferObject_SectorIndex;
  var FileObject;
  var FileObject_acFullFileName;
  var FileObject_SizeEncrypted;
  var Expression;
  var FS_Lock;
  var FS_FAT_CONFIG;
  var FS_EFS_CONFIG;
  var FS_EFS_AllocDirEntryBuffer_Addr;
  var FS_CONTEXT;
  var FS_CONTEXT_IsWriteVerificationEnabled;
  var FS_CONTEXT_IsEOFErrorSuppressed;
  var FS_CONTEXT_IsPOSIXSupported;
  var FS_CONTEXT_EFS_pUnicodeConv;
  var FS_CONTEXT_FAT_MaintainFATCopy;
  var FS_CONTEXT_FAT_PermitROFileMove;
  var FS_CONTEXT_EFS_IsCaseSensitive;
  var CASE_INFO;
  var JOURNAL_STATUS;
  var JOURNAL_STATUS_StatCounters;
  var JOURNAL_INST;
  var JOURNAL_INST_paIsSectorFree;
  var JOURNAL_INST_NumBitsSectorIndex;
  var JOURNAL_INST_pSC;
  var _WriteEmptyFATSectors;
  var FS_INT_DATA_FAT;
  var FS_INT_DATA_FAT_NumAdjClusters;
  var _LoadShortNameSB;
  var _IsEmptyDirEntry;
  var _WriteEmptyEFSSectors;
  var FS_INT_DATA_EFS;
  var FS_INT_DATA_EFS_NumAdjClusters;
  var FS_STORAGE_GetCounters;
  var FS_DEVICE_DATA;
  var FS_DEVICE_DATA_pfOnDeviceActivity;
  var NAND_UNI_INST;
  var FAT_LFN_API;
  
  _UpdateTargetInfoFsGlobal();
  if (_Cache_FS_Global != undefined) {
    _Feature_EmFilePresent = true;
  }
  if (_Feature_EmFilePresent == true) {
    Expression    = "*(FS_FAT_CONFIG*)0";
    FS_FAT_CONFIG = Debug.evaluate(Expression);
    if (FS_FAT_CONFIG != undefined) {
      _Feature_FS_SUPPORT_FAT = true;
    }
    Expression    = "*(FS_EFS_CONFIG*)0";
    FS_EFS_CONFIG = Debug.evaluate(Expression);
    if (FS_EFS_CONFIG != undefined) {
      _Feature_FS_SUPPORT_EFS = true;
    }
    FS_Global_pFirstOnExit = _Cache_FS_Global["pFirstOnExit"];
    if (FS_Global_pFirstOnExit != undefined) {
      _Feature_FS_SUPPORT_DEINIT = true;
    }
    FS_Global_FileBufferSize = _Cache_FS_Global["FileBufferSize"];
    if (FS_Global_FileBufferSize != undefined) {
      _Feature_FS_SUPPORT_FILE_BUFFER = true;
    }
    FS_Global_Profile = _Cache_FS_Global["Profile"];
    if (FS_Global_Profile != undefined) {
      _Feature_FS_SUPPORT_PROFILE = true;
    }
    FS_Global_FirstVolume = _Cache_FS_Global["FirstVolume"];
    if (FS_Global_FirstVolume != undefined) {
      FS_Global_FirstVolume_pFS_API = FS_Global_FirstVolume["pFS_API"];
      if (FS_Global_FirstVolume_pFS_API != undefined) {
        _Feature_FS_SUPPORT_MULTIPLE_FS = true;
      }
      FS_Global_FirstVolume_FreeSector = FS_Global_FirstVolume["FreeSector"];
      if (FS_Global_FirstVolume_FreeSector != undefined) {
        _Feature_FS_SUPPORT_FREE_SECTOR = true;
      }
      FS_Global_FirstVolume_Partition = FS_Global_FirstVolume["Partition"];
      if(FS_Global_FirstVolume_Partition != undefined) {
        FS_Global_FirstVolume_Partition_Device = FS_Global_FirstVolume_Partition["Device"];
        if(FS_Global_FirstVolume_Partition_Device != undefined) {
          FS_Global_FirstVolume_Partition_Device_Data = FS_Global_FirstVolume_Partition_Device["Data"];
          if(FS_Global_FirstVolume_Partition_Device_Data != undefined) {
            FS_Global_FirstVolume_Partition_Device_Data_JournalData = FS_Global_FirstVolume_Partition_Device_Data["JournalData"];
            if(FS_Global_FirstVolume_Partition_Device_Data_JournalData != undefined) {
              _Feature_FS_SUPPORT_JOURNAL = true;
            }
            FS_Global_FirstVolume_Partition_Device_Data_pCacheData = FS_Global_FirstVolume_Partition_Device_Data["pCacheData"];
            if(FS_Global_FirstVolume_Partition_Device_Data_pCacheData != undefined) {
              _Feature_FS_SUPPORT_CACHE = true;
            }
            FS_Global_FirstVolume_Partition_Device_Data_pfSetBusyLED = FS_Global_FirstVolume_Partition_Device_Data["pfSetBusyLED"];
            if(FS_Global_FirstVolume_Partition_Device_Data_pfSetBusyLED != undefined) {
              _Feature_FS_SUPPORT_BUSY_LED = true;
            }
            FS_Global_FirstVolume_Partition_Device_Data_pfMemCheck = FS_Global_FirstVolume_Partition_Device_Data["pfMemCheck"];
            if(FS_Global_FirstVolume_Partition_Device_Data_pfMemCheck != undefined) {
              _Feature_FS_SUPPORT_CHECK_MEMORY = true;
            }
            FS_Global_FirstVolume_Partition_Device_Data_pfOnDeviceActivity = FS_Global_FirstVolume_Partition_Device_Data["pfOnDeviceActivity"];
            if(FS_Global_FirstVolume_Partition_Device_Data_pfOnDeviceActivity != undefined) {
              _Feature_FS_STORAGE_SUPPORT_DEVICE_ACTIVITY = true;
            }
          }
        }
      }
      FS_Global_FirstVolume_acJournalFileName = FS_Global_FirstVolume["acJournalFileName"];
      if (FS_Global_FirstVolume_acJournalFileName != undefined) {
        _Feature_FS_SUPPORT_JOURNAL_FILENAME = true;
      }
      FS_Global_FirstVolume_acAlias = FS_Global_FirstVolume["acAlias"];
      FS_Global_FirstVolume_sAlias = FS_Global_FirstVolume["sAlias"];
      if (FS_Global_FirstVolume_acAlias != undefined) {
        _Feature_FS_SUPPORT_VOLUME_ALIAS = "Enabled, ownAliasName";
      } else if (FS_Global_FirstVolume_sAlias != undefined){
        _Feature_FS_SUPPORT_VOLUME_ALIAS = "Enabled, othersAliasName";
      }
      FS_Global_FirstVolume_FSInfo         = FS_Global_FirstVolume["FSInfo"];
      if (FS_Global_FirstVolume_FSInfo != undefined) {
        FS_Global_FirstVolume_FSInfo_EFSInfo = FS_Global_FirstVolume_FSInfo["EFSInfo"];
        if (FS_Global_FirstVolume_FSInfo_EFSInfo != undefined) {
          FS_Global_FirstVolume_FSInfo_EFSInfo_IsNumFreeClustersInvalidated = FS_Global_FirstVolume_FSInfo_EFSInfo["IsNumFreeClustersInvalidated"];
          if (FS_Global_FirstVolume_FSInfo_EFSInfo_IsNumFreeClustersInvalidated != undefined) {
            if (_Feature_FS_SUPPORT_EFS == true) {
              _Feature_FS_EFS_SUPPORT_STATUS_SECTOR = true;
            }
          }
          FS_Global_FirstVolume_FSInfo_EFSInfo_FreeClusterCache = FS_Global_FirstVolume_FSInfo_EFSInfo["FreeClusterCache"];
          if (FS_Global_FirstVolume_FSInfo_EFSInfo_FreeClusterCache != undefined) {
            if (_Feature_FS_SUPPORT_EFS == true) {
              _Feature_FS_EFS_SUPPORT_FREE_CLUSTER_CACHE = true;
            }
          }
        }
        FS_Global_FirstVolume_FSInfo_FATInfo = FS_Global_FirstVolume_FSInfo["FATInfo"];
        if (FS_Global_FirstVolume_FSInfo_FATInfo != undefined) {
          FS_Global_FirstVolume_FSInfo_FATInfo_FSInfoSector = FS_Global_FirstVolume_FSInfo_FATInfo["FSInfoSector"];
          if (FS_Global_FirstVolume_FSInfo_FATInfo_FSInfoSector != undefined) {
            if (_Feature_FS_SUPPORT_FAT == true) {
              _Feature_FS_FAT_USE_FSINFO_SECTOR = true;
            }
          }
        }
      }
      FS_Global_FirstVolume_FSInfo_FATInfo_IsDirty = FS_Global_FirstVolume_FSInfo_FATInfo["IsDirty"];
      if (FS_Global_FirstVolume_FSInfo_FATInfo_IsDirty != undefined) {
        if (_Feature_FS_SUPPORT_FAT == true) {
          _Feature_FS_FAT_UPDATE_DIRTY_FLAG = true;
        }
      }
      FS_Global_FirstVolume_FSInfo_FATInfo_FreeClusterCache = FS_Global_FirstVolume_FSInfo_FATInfo["FreeClusterCache"];
      if (FS_Global_FirstVolume_FSInfo_FATInfo_FreeClusterCache != undefined) {
        if (_Feature_FS_SUPPORT_FAT == true) {
          _Feature_FS_FAT_SUPPORT_FREE_CLUSTER_CACHE = true;
        }
      }
    }
    FS_Global_MemManager = _Cache_FS_Global["MemManager"];
    if (FS_Global_MemManager != undefined) {
      FS_Global_MemManager_pfAlloc = FS_Global_MemManager["pfAlloc"];
      if (FS_Global_MemManager_pfAlloc != undefined) {
        _Feature_FS_SUPPORT_EXT_MEM_MANAGER = true;
      }
    }
    _UpdateTargetInfoSectorBufferObject();
    SectorBufferObject = _Cache_SectorBufferObject;
    if (SectorBufferObject != undefined) {
       SectorBufferObject_SectorIndex = SectorBufferObject["SectorIndex"];
       if (SectorBufferObject_SectorIndex != undefined) {
         _Feature_FS_SUPPORT_SECTOR_BUFFER_CACHE = true;
      }
    }
    _UpdateTargetInfoFileObject();
    FileObject = _Cache_FileObject;
    if (FileObject != undefined) {
       FileObject_acFullFileName = FileObject["acFullFileName"];
       if (FileObject_acFullFileName != undefined) {
         _Feature_FS_MULTI_HANDLE_SAFE = true;
      }
       FileObject_SizeEncrypted = FileObject["SizeEncrypted"];
       if (FileObject_SizeEncrypted != undefined) {
         _Feature_FS_SUPPORT_ENCRYPTION = true;
      }
    }
    Expression = "FS_Lock";
    FS_Lock    = Debug.evaluate(Expression);
    if (FS_Lock != undefined) {
      _Feature_FS_OS_LOCKING = "FS_OS_LOCKING_API"
    } else {
      Expression    = "FS_LockVolume";
      FS_LockVolume = Debug.evaluate(Expression);
      if (FS_LockVolume != undefined) {
        _Feature_FS_OS_LOCKING = "FS_OS_LOCKING_DRIVER";
      } else {
        _Feature_FS_OS_LOCKING = "FS_OS_LOCKING_NONE";
      }
    }
    Expression                      = "&FS_EFS_AllocDirEntryBuffer";
    FS_EFS_AllocDirEntryBuffer_Addr = Debug.evaluate(Expression);
    if (FS_EFS_AllocDirEntryBuffer_Addr != undefined) {
      if (_Feature_FS_SUPPORT_EFS == true) {
        _Feature_FS_EFS_SUPPORT_DIRENTRY_BUFFERS = true;
      }
    }
    Expression = "*(FS_CONTEXT*)0";
    FS_CONTEXT = Debug.evaluate(Expression);
    if (FS_CONTEXT != undefined) {
      FS_CONTEXT_IsWriteVerificationEnabled = FS_CONTEXT["IsWriteVerificationEnabled"];
      if (FS_CONTEXT_IsWriteVerificationEnabled != undefined) {
        _Feature_FS_VERIFY_WRITE = true;
      }
      FS_CONTEXT_IsEOFErrorSuppressed = FS_CONTEXT["IsEOFErrorSuppressed"];
      if (FS_CONTEXT_IsEOFErrorSuppressed != undefined) {
        _Feature_FS_SUPPRESS_EOF_ERROR = true;
      }
      FS_CONTEXT_IsPOSIXSupported = FS_CONTEXT["IsPOSIXSupported"];
      if (FS_CONTEXT_IsPOSIXSupported != undefined) {
        _Feature_FS_SUPPORT_POSIX = true;
      }
      FS_CONTEXT_EFS_pUnicodeConv = FS_CONTEXT["EFS_pUnicodeConv"];
      if (FS_CONTEXT_EFS_pUnicodeConv != undefined) {
        _Feature_FS_SUPPORT_FILE_NAME_ENCODING = true;
      }
      FS_CONTEXT_FAT_MaintainFATCopy = FS_CONTEXT["FAT_MaintainFATCopy"];
      if (FS_CONTEXT_FAT_MaintainFATCopy != undefined) {
        _Feature_FS_MAINTAIN_FAT_COPY = true;
      }
      FS_CONTEXT_FAT_PermitROFileMove = FS_CONTEXT["FAT_PermitROFileMove"];
      if (FS_CONTEXT_FAT_PermitROFileMove != undefined) {
        if (_Feature_FS_SUPPORT_FAT == true) {
          _Feature_FS_FAT_PERMIT_RO_FILE_MOVE = true;
        }
      }
      FS_CONTEXT_EFS_IsCaseSensitive = FS_CONTEXT["EFS_IsCaseSensitive"];
      if (FS_CONTEXT_EFS_IsCaseSensitive != undefined) {
        if (_Feature_FS_SUPPORT_EFS == true) {
          _Feature_FS_EFS_CASE_SENSITIVE = true;
        }
      }
    }
    Expression = "&FS_FAT_StoreShortNamePartialMB()";
    FS_FAT_StoreShortNamePartialMB = Debug.evaluate(Expression);
    if (FS_FAT_StoreShortNamePartialMB != undefined) {
      _Feature_FS_SUPPORT_MBCS = true;
    }
    Expression = "*(CASE_INFO*)0";
    CASE_INFO  = Debug.evaluate(Expression);
    if (CASE_INFO != undefined) {
      _Feature_FS_SUPPORT_EXT_ASCII = true;
    }
    Expression = "*(JOURNAL_STATUS*)0";
    JOURNAL_STATUS  = Debug.evaluate(Expression);
    if (JOURNAL_STATUS != undefined) {
      JOURNAL_STATUS_StatCounters = JOURNAL_STATUS["StatCounters"];
      if (JOURNAL_STATUS_StatCounters != undefined) {
        _Feature_FS_JOURNAL_ENABLE_STATS = true;
      }
    }
    Expression = "*(JOURNAL_INST*)0";
    JOURNAL_INST  = Debug.evaluate(Expression);
    if (JOURNAL_INST != undefined) {
      JOURNAL_INST_paIsSectorFree = JOURNAL_STATUS["paIsSectorFree"];
      if (JOURNAL_INST_paIsSectorFree != undefined) {
        _Feature_FS_JOURNAL_SUPPORT_FREE_SECTOR = true;
      }
      JOURNAL_INST_NumBitsSectorIndex = JOURNAL_INST["NumBitsSectorIndex"];
      if (JOURNAL_INST_NumBitsSectorIndex != undefined) {
        _Feature_FS_JOURNAL_SUPPORT_FAST_SECTOR_SEARCH = true;
      }
      JOURNAL_INST_pSC = JOURNAL_INST["pSC"];
      if (JOURNAL_INST_pSC != undefined) {
        _Feature_FS_JOURNAL_OPTIMIZE_SPACE_USAGE = true;
      }
    }
    Expression = "_WriteEmptyFATSectors()";
    _WriteEmptyFATSectors = Debug.evaluate(Expression);
    if (_WriteEmptyFATSectors != undefined) {
      if (_Feature_FS_SUPPORT_FAT == true) {
        _Feature_FS_FAT_OPTIMIZE_DELETE = true;
      }
    }
    Expression = "*(FS_INT_DATA_FAT*)0";
    FS_INT_DATA_FAT = Debug.evaluate(Expression);
    if (FS_INT_DATA_FAT != undefined) {
      FS_INT_DATA_FAT_NumAdjClusters = FS_INT_DATA_FAT["NumAdjClusters"];
      if (FS_INT_DATA_FAT_NumAdjClusters != undefined) {
        if (_Feature_FS_SUPPORT_FAT == true) {
          _Feature_FS_FAT_OPTIMIZE_LINEAR_ACCESS = true;
        }
      }
    }
    Expression = "_LoadShortNameSB()";
    _LoadShortNameSB = Debug.evaluate(Expression);
    if (_LoadShortNameSB != undefined) {
      if (_Feature_FS_SUPPORT_FAT == true) {
        _Feature_FS_FAT_LFN_LOWER_CASE_SHORT_NAMES = true;
      }
    }
    Expression = "_IsEmptyDirEntry()";
    _IsEmptyDirEntry = Debug.evaluate(Expression);
    if (_IsEmptyDirEntry != undefined) {
      if (_Feature_FS_SUPPORT_FAT == true) {
        _Feature_FS_FAT_CHECK_UNUSED_DIR_ENTRIES = true;
      }
    }
    Expression = "_WriteEmptyEFSSectors()";
    _WriteEmptyEFSSectors = Debug.evaluate(Expression);
    if (_WriteEmptyEFSSectors != undefined) {
      if (_Feature_FS_SUPPORT_EFS == true) {
        _Feature_FS_EFS_OPTIMIZE_DELETE = true;
      }
    }
    Expression = "*(FS_INT_DATA_EFS*)0";
    FS_INT_DATA_EFS = Debug.evaluate(Expression);
    if (FS_INT_DATA_EFS != undefined) {
      FS_INT_DATA_EFS_NumAdjClusters = FS_INT_DATA_EFS["NumAdjClusters"];
      if (FS_INT_DATA_EFS_NumAdjClusters != undefined) {
        if (_Feature_FS_SUPPORT_EFS == true) {
          _Feature_FS_EFS_OPTIMIZE_LINEAR_ACCESS = true;
        }
      }
    }
    Expression = "FS_STORAGE_GetCounters()";
    FS_STORAGE_GetCounters = Debug.evaluate(Expression);
    if (FS_STORAGE_GetCounters != undefined) {
      _Feature_FS_STORAGE_ENABLE_STAT_COUNTERS = true;
    }
    Expression = "*(FS_DEVICE_DATA*)0";
    FS_DEVICE_DATA = Debug.evaluate(Expression);
    if (FS_DEVICE_DATA != undefined) {
      FS_DEVICE_DATA_pfOnDeviceActivity = FS_DEVICE_DATA["pfOnDeviceActivity"];
      if (FS_DEVICE_DATA_pfOnDeviceActivity != undefined) {
        _Feature_FS_STORAGE_SUPPORT_DEVICE_ACTIVITY = true;
      }
    }
    Expression = "FS_OS_SetType()";
    FS_OS_SetType = Debug.evaluate(Expression);
    if (FS_OS_SetType != undefined) {
      _Feature_FS_OS_SUPPORT_RUNTIME_CONFIG = true;
    }
    Expression = "*(NAND_UNI_INST*)0";
    NAND_UNI_INST = Debug.evaluate(Expression);
    if (NAND_UNI_INST != undefined) {
      _Feature_FS_NAND_ENABLE_STATS = true;
    }
    Expression = "FAT_LFN_API";
    FAT_LFN_API = Debug.evaluate(Expression);
    if (FAT_LFN_API != undefined) {
      if (_Feature_FS_SUPPORT_FAT == true) {
        _Feature_FS_FAT_SUPPORT_LFN = true;
      }
    }
  }
}

/*********************************************************************
*
* Page Script Info
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageScriptInfoColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageScriptInfoColHeader () {
  var Result;

  Result = ["Script Info"];
  return Result;
}

/*********************************************************************
*
*       _GetPageScriptInfoRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPageScriptInfoRow(RowIdx) {
  var Result;

  switch (RowIdx) {
    case 0:
      Result = ["emFile plug-in V1.0.0"];
      break;
    case 1:
      Result = ["Last modified: 2022-08-24"];
      break;
    case 2:
      Result = ["Tested on emFile V5.18.00"];
      break;
    case 3:
      Result = ["Copyright (c) 2022 SEGGER Microcontroller GmbH"];
      break;
    default:
      Result = null;
      break;
  }
  return Result;
}

/*********************************************************************
*
* Page Features
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageFeaturesColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageFeaturesColHeader () {
  var Result;

  Result = ["Item", "State"];
  return Result;
}

/*********************************************************************
*
*       _GetPageFeaturesMemberRow()
*
*  Functions description:
*    Translates the values of a member of the respective data structure
*    into strings and composes the row content.
*/
function _GetPageFeaturesMemberRow(ItemName, Value, ToStrFunction) {
  var Result;
  var sValue;

  Result = [];
  Result.push(ItemName);
  sValue = ToStrFunction(Value);
  Result.push(sValue);
  return Result;
}

/*********************************************************************
*
*       _GetPageFeaturesRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPageFeaturesRow(RowIdx) {
  var Result;

  switch (RowIdx) {
    case 0:
      _UpdateFeatureFlags();

      Result = _GetPageFeaturesMemberRow("emFile detected", _Feature_EmFilePresent, _ToStrBoolYesNo);
      break;
    case 1:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_DEINIT", _Feature_FS_SUPPORT_DEINIT, _ToStrBoolEnabledDisable);
      break;
    case 2:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_FILE_BUFFER", _Feature_FS_SUPPORT_FILE_BUFFER, _ToStrBoolEnabledDisable);
      break;
    case 3:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_SECTOR_BUFFER_CACHE", _Feature_FS_SUPPORT_SECTOR_BUFFER_CACHE, _ToStrBoolEnabledDisable);
      break;
    case 4:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_PROFILE", _Feature_FS_SUPPORT_PROFILE, _ToStrBoolEnabledDisable);
      break;
    case 5:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_MULTIPLE_FS", _Feature_FS_SUPPORT_MULTIPLE_FS, _ToStrBoolEnabledDisable);
      break;
    case 6:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_FREE_SECTOR", _Feature_FS_SUPPORT_FREE_SECTOR, _ToStrBoolEnabledDisable);
      break;
    case 7:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_JOURNAL", _Feature_FS_SUPPORT_JOURNAL, _ToStrBoolEnabledDisable);
      break;
    case 8:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_CACHE", _Feature_FS_SUPPORT_CACHE, _ToStrBoolEnabledDisable);
      break;
    case 9:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_BUSY_LED", _Feature_FS_SUPPORT_BUSY_LED, _ToStrBoolEnabledDisable);
      break;
    case 10:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_CHECK_MEMORY", _Feature_FS_SUPPORT_CHECK_MEMORY, _ToStrBoolEnabledDisable);
      break;
    case 11:
      Result = _GetPageFeaturesMemberRow("FS_STORAGE_SUPPORT_DEVICE_ACTIVITY", _Feature_FS_STORAGE_SUPPORT_DEVICE_ACTIVITY, _ToStrBoolEnabledDisable);
      break;
    case 12:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_JOURNAL (Filename)", _Feature_FS_SUPPORT_JOURNAL_FILENAME, _ToStrBoolEnabledDisable);
      break;
    case 13:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_VOLUME_ALIAS", _Feature_FS_SUPPORT_VOLUME_ALIAS, _ToStrOrEmptyTag);
      break;
    case 14:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_EXT_MEM_MANAGER", _Feature_FS_SUPPORT_EXT_MEM_MANAGER, _ToStrBoolEnabledDisable);
      break;
    case 15:
      Result = _GetPageFeaturesMemberRow("FS_MULTI_HANDLE_SAFE", _Feature_FS_MULTI_HANDLE_SAFE, _ToStrBoolEnabledDisable);
      break;
    case 16:
      Result = _GetPageFeaturesMemberRow("FS_EFS_SUPPORT_STATUS_SECTOR", _Feature_FS_EFS_SUPPORT_STATUS_SECTOR, _ToStrBoolEnabledDisable);
      break;
    case 17:
      Result = _GetPageFeaturesMemberRow("FS_EFS_SUPPORT_FREE_CLUSTER_CACHE", _Feature_FS_EFS_SUPPORT_FREE_CLUSTER_CACHE, _ToStrBoolEnabledDisable);
      break;
    case 18:
      Result = _GetPageFeaturesMemberRow("FS_FAT_USE_FSINFO_SECTOR", _Feature_FS_FAT_USE_FSINFO_SECTOR, _ToStrBoolEnabledDisable);
      break;
    case 19:
      Result = _GetPageFeaturesMemberRow("FS_FAT_UPDATE_DIRTY_FLAG", _Feature_FS_FAT_UPDATE_DIRTY_FLAG, _ToStrBoolEnabledDisable);
      break;
    case 20:
      Result = _GetPageFeaturesMemberRow("FS_FAT_SUPPORT_FREE_CLUSTER_CACHE", _Feature_FS_FAT_SUPPORT_FREE_CLUSTER_CACHE, _ToStrBoolEnabledDisable);
      break;
    case 21:
      Result = _GetPageFeaturesMemberRow("FS_EFS_SUPPORT_DIRENTRY_BUFFERS", _Feature_FS_EFS_SUPPORT_DIRENTRY_BUFFERS, _ToStrBoolEnabledDisable);
      break;
    case 22:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_ENCRYPTION", _Feature_FS_SUPPORT_ENCRYPTION, _ToStrBoolEnabledDisable);
      break;
    case 23:
      Result = _GetPageFeaturesMemberRow("FS_OS_LOCKING", _Feature_FS_OS_LOCKING, _ToStrOrEmptyTag);
      break;
    case 24:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_FAT", _Feature_FS_SUPPORT_FAT, _ToStrBoolEnabledDisable);
      break;
    case 25:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_EFS", _Feature_FS_SUPPORT_EFS, _ToStrBoolEnabledDisable);
      break;
    case 26:
      Result = _GetPageFeaturesMemberRow("FS_VERIFY_WRITE", _Feature_FS_VERIFY_WRITE, _ToStrBoolEnabledDisable);
      break;
    case 27:
      Result = _GetPageFeaturesMemberRow("FS_SUPPRESS_EOF_ERROR", _Feature_FS_SUPPRESS_EOF_ERROR, _ToStrBoolEnabledDisable);
      break;
    case 28:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_POSIX", _Feature_FS_SUPPORT_POSIX, _ToStrBoolEnabledDisable);
      break;
    case 29:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_MBCS", _Feature_FS_SUPPORT_MBCS, _ToStrBoolEnabledDisable);
      break;
    case 30:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_EXT_ASCII", _Feature_FS_SUPPORT_EXT_ASCII, _ToStrBoolEnabledDisable);
      break;
    case 31:
      Result = _GetPageFeaturesMemberRow("FS_SUPPORT_FILE_NAME_ENCODING", _Feature_FS_SUPPORT_FILE_NAME_ENCODING, _ToStrBoolEnabledDisable);
      break;
    case 32:
      Result = _GetPageFeaturesMemberRow("FS_JOURNAL_ENABLE_STATS", _Feature_FS_JOURNAL_ENABLE_STATS, _ToStrBoolEnabledDisable);
      break;
    case 33:
      Result = _GetPageFeaturesMemberRow("FS_JOURNAL_SUPPORT_FREE_SECTOR", _Feature_FS_JOURNAL_SUPPORT_FREE_SECTOR, _ToStrBoolEnabledDisable);
      break;
    case 34:
      Result = _GetPageFeaturesMemberRow("FS_JOURNAL_SUPPORT_FAST_SECTOR_SEARCH", _Feature_FS_JOURNAL_SUPPORT_FAST_SECTOR_SEARCH, _ToStrBoolEnabledDisable);
      break;
    case 35:
      Result = _GetPageFeaturesMemberRow("FS_JOURNAL_OPTIMIZE_SPACE_USAGE", _Feature_FS_JOURNAL_OPTIMIZE_SPACE_USAGE, _ToStrBoolEnabledDisable);
      break;
    case 36:
      Result = _GetPageFeaturesMemberRow("FS_MAINTAIN_FAT_COPY", _Feature_FS_MAINTAIN_FAT_COPY, _ToStrBoolEnabledDisable);
      break;
    case 37:
      Result = _GetPageFeaturesMemberRow("FS_FAT_OPTIMIZE_DELETE", _Feature_FS_FAT_OPTIMIZE_DELETE, _ToStrBoolEnabledDisable);
      break;
    case 38:
      Result = _GetPageFeaturesMemberRow("FS_FAT_OPTIMIZE_LINEAR_ACCESS", _Feature_FS_FAT_OPTIMIZE_LINEAR_ACCESS, _ToStrBoolEnabledDisable);
      break;
    case 39:
      Result = _GetPageFeaturesMemberRow("FS_FAT_PERMIT_RO_FILE_MOVE", _Feature_FS_FAT_PERMIT_RO_FILE_MOVE, _ToStrBoolEnabledDisable);
      break;
    case 40:
      Result = _GetPageFeaturesMemberRow("FS_FAT_LFN_LOWER_CASE_SHORT_NAMES", _Feature_FS_FAT_LFN_LOWER_CASE_SHORT_NAMES, _ToStrBoolEnabledDisable);
      break;
    case 41:
      Result = _GetPageFeaturesMemberRow("FS_FAT_CHECK_UNUSED_DIR_ENTRIES", _Feature_FS_FAT_CHECK_UNUSED_DIR_ENTRIES, _ToStrBoolEnabledDisable);
      break;
    case 42:
      Result = _GetPageFeaturesMemberRow("FS_EFS_OPTIMIZE_DELETE", _Feature_FS_EFS_OPTIMIZE_DELETE, _ToStrBoolEnabledDisable);
      break;
    case 43:
      Result = _GetPageFeaturesMemberRow("FS_EFS_CASE_SENSITIVE", _Feature_FS_EFS_CASE_SENSITIVE, _ToStrBoolEnabledDisable);
      break;
    case 44:
      Result = _GetPageFeaturesMemberRow("FS_EFS_OPTIMIZE_LINEAR_ACCESS", _Feature_FS_EFS_OPTIMIZE_LINEAR_ACCESS, _ToStrBoolEnabledDisable);
      break;
    case 45:
      Result = _GetPageFeaturesMemberRow("FS_STORAGE_ENABLE_STAT_COUNTERS", _Feature_FS_STORAGE_ENABLE_STAT_COUNTERS, _ToStrBoolEnabledDisable);
      break;
    case 46:
      Result = _GetPageFeaturesMemberRow("FS_STORAGE_SUPPORT_DEVICE_ACTIVITY", _Feature_FS_STORAGE_SUPPORT_DEVICE_ACTIVITY, _ToStrBoolEnabledDisable);
      break;
    case 47:
      Result = _GetPageFeaturesMemberRow("FS_OS_SUPPORT_RUNTIME_CONFIG", _Feature_FS_OS_SUPPORT_RUNTIME_CONFIG, _ToStrBoolEnabledDisable);
      break;
    case 48:
      Result = _GetPageFeaturesMemberRow("FS_NAND_ENABLE_STATS", _Feature_FS_NAND_ENABLE_STATS, _ToStrBoolEnabledDisable);
      break;
    case 49:
      Result = _GetPageFeaturesMemberRow("FS_FAT_SUPPORT_LFN", _Feature_FS_FAT_SUPPORT_LFN, _ToStrBoolEnabledDisable);
      break;
    case 50:
      _LfnSupportDetection();
      Result = _GetPageFeaturesMemberRow("Long filename support", _Feature_LongFilenameSupport, _ToStr);
      break;
    default:
      Result = null;
      break;
  }
  return Result;
}

/*********************************************************************
*
* Page Global
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageGlobalColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageGlobalColHeader () {
  var Result;

  _UpdateFeatureFlags();
  Result = ["Item"];
  if (_Feature_EmFilePresent == true) {
    Result.push("Value");
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageGlobalMemberRow()
*
*  Functions description:
*    Translates the values of a member of the respective data structure
*    into strings and composes the row content.
*/
function _GetPageGlobalMemberRow(ItemName, Member, ToStrFunction) {
  var Result;
  var Value;
  var sValue;

  Result = [];
  Result.push(ItemName);
  _UpdateFeatureFlags();
  if (_Feature_EmFilePresent == true) {
    Value  = _Cache_FS_Global[Member];
    sValue = ToStrFunction(Value);
    Result.push (sValue);
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageGlobalCopyrightMemberRow()
*
*  Functions description:
*    Translates the values of a member of the respective data structure
*    into strings and composes the row content.
*/
function _GetPageGlobalCopyrightMemberRow(ItemName, Member, ToStrFunction) {
  var Result;
  var sValue;

  Result = [];
  Result.push(ItemName);
  _UpdateFeatureFlags();
  if (_Feature_EmFilePresent == true) {
    sValue  = _Cache_FS_Global_Copyright;
    Result.push (sValue);
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageGlobalRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPageGlobalRow(RowIdx) {
  var Result;

  switch (RowIdx) {
    case 0:
      _UpdateTargetInfoFsGlobal();
      Result = _GetPageGlobalMemberRow("Number of sector buffers", "NumSectorBuffers", _ToStr);
      break;
    case 1:
      Result = _GetPageGlobalMemberRow("Max sector size", "MaxSectorSize", _ToStrByteSuffix);
      break;
    case 2:
      Result = _GetPageGlobalMemberRow("Is inited", "IsInited", _ToStrBoolYesNo);
      break;
    case 3:;
      Result = _GetPageGlobalMemberRow("Is storage inited", "IsStorageInited", _ToStrBoolYesNo);
      break;
    case 4:
      Result = _GetPageGlobalMemberRow("Number of volumes", "NumVolumes", _ToStr);
      break;
    case 5:
      Result = _GetPageGlobalMemberRow("Write mode", "WriteMode", _ToStrWriteMode);
      break;
    case 6:
      if (_Feature_FS_SUPPORT_FILE_BUFFER == true) {
        Result = _GetPageGlobalMemberRow("File buffer size", "FileBufferSize", _ToStrByteSuffix);
      } else {
        Result = _GetFeatureNotEnabledVRow("File buffer size", 1);
      }
      break;
    case 7:
      if (_Feature_FS_SUPPORT_FILE_BUFFER == true) {
        Result = _GetPageGlobalMemberRow("File buffer flags", "FileBufferFlags", _ToStrFileBufferFlags);
      } else {
        Result = _GetFeatureNotEnabledVRow("File buffer flags", 1);
      }
      break;
    case 8:
        _UpdateTargetInfoFsGlobalCopyright();
        Result = _GetPageGlobalCopyrightMemberRow("Copyright");
      break;
    default:
      Result = null;
      break;
  }
  return Result;
}

/*********************************************************************
*
* Page Memory Manager
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageMemManagerColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageMemManagerColHeader () {
  var Result;

  Result = ["Item"];
  _UpdateFeatureFlags();
  if (_Feature_EmFilePresent == true) {
    Result.push("Value");
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageMemManagerMemberRow()
*
*  Functions description:
*    Translates the values of a member of the respective data structure
*    into strings and composes the row content.
*/
function _GetPageMemManagerMemberRow(ItemName, Member, ToStrFunction) {
  var Result;
  var MemManagerInfo;
  var Value;
  var sValue;
  
  Result = [];
  Result.push(ItemName);
  _UpdateFeatureFlags();
  if (_Feature_EmFilePresent == true) {
    MemManagerInfo = _Cache_FS_Global["MemManager"];
    Value          = MemManagerInfo[Member];
    sValue         = ToStrFunction(Value);
    Result.push (sValue);
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageMemManagerRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPageMemManagerRow(RowIdx) {
  var Result;

  switch (RowIdx) {
    case 0:
      _UpdateTargetInfoFsGlobal();

      Result = _GetPageMemManagerMemberRow("Allocated bytes", "NumBytesAllocated", _ToStrByteSuffix);
      break;
    case 1:
      if (_Feature_FS_SUPPORT_EXT_MEM_MANAGER == false) {
        Result = _GetPageMemManagerMemberRow("Total bytes", "NumBytesTotal", _ToStrByteSuffix);
      } else {
        Result = _GetFeatureNotEnabledVRow("Total bytes", 1);
      }
      break;
    case 2:
      if (_Feature_FS_SUPPORT_DEINIT == true) {
        Result = _GetPageMemManagerMemberRow("Allocated blocks", "NumBlocksAllocated", _ToStr);
      } else {
        Result = _GetFeatureNotEnabledVRow("Allocated blocks", 1);
      }
      break;
    case 3:
      if (_Feature_FS_SUPPORT_EXT_MEM_MANAGER == true) {
        Result = _GetPageMemManagerMemberRow("Function alloc()", "pfAlloc", _ToStrFuncNameOrUnknownTag);
      } else {
        Result = _GetFeatureNotEnabledVRow("Function alloc()", 1);
      }
      break;
    case 4:
      if (_Feature_FS_SUPPORT_EXT_MEM_MANAGER == true) {
        Result = _GetPageMemManagerMemberRow("Function free()", "pfFree", _ToStrFuncNameOrUnknownTag);
      } else {
        Result = _GetFeatureNotEnabledVRow("Function free()", 1);
      }
      break;
    default:
      Result = null;
      break;
  }
  return Result;
}

/*********************************************************************
*
* Page File Handles
*
*********************************************************************/

/*********************************************************************
*
*       _ToStrFileHandleFlagsWRCBA()
*
*  Functions description:
*    Creates a case-sensitive display of the file buffer flags.
*/
function _ToStrFileHandleFlagsWRCBA(Value) {
  var Result;
  var HexVal;

  Result = "";
  if ((Value & 0x10) == 0) {
    Result +="w";
  } else {
    Result +="W";
  }
  if ((Value & 0x08) == 0) {
    Result +="r";
  } else {
    Result +="R";
  }
  if ((Value & 0x04) == 0) {
    Result +="c";
  } else {
    Result +="C";
  }
  if ((Value & 0x02) == 0) {
    Result +="b";
  } else {
    Result +="B";
  }
  if ((Value & 0x01) == 0) {
    Result +="a";
  } else {
    Result +="A";
  }
  HexVal = _ToStrHex(Value);
  Result += " (" + HexVal + ")";
  return Result;
}

/*********************************************************************
*
*       _ToStrFileHandleAccessMode()
*
*  Functions description:
*    Creates a case-sensitive display of the file buffer flags.
*    Table taken from _aAccessMode[] in FS_Misc.c.
*/
function _ToStrFileHandleAccessMode(Value) {
  var Result;
  var HexVal;

  Result = "";
  switch (Value) {
    case 0x08: 
      Result = "\"r\"";
      break;
    case 0x0A: 
      Result = "\"rb\"";
      break;
    case 0x14:
      Result = "\"w\"";
      break;
    case 0x15:
      Result = "\"a\"";
      break;
    case 0x16:
      Result = "\"wb\"";
      break;
    case 0x17:
      Result = "\"ab\"";
      break;
    case 0x18:
      Result = "\"r+\"";
      break;
    case 0x1A:
      Result = "\"r+b\"";
      break;
    case 0x1C:
      Result = "\"w+\"";
      break;
    case 0x1D:
      Result = "\"a+\"";
      break;
    case 0x1E:
      Result = "\"w+b\"";
      break;
    case 0x1F:
      Result = "\"a+b\"";
      break;
    default:
      Result = "<unknown access mode: " + _ToStrHex(Value) + ">";
      break;
  }
  return Result;
}

/*********************************************************************
*
*       _UpdateTargetInfoFileHandles()
*
*  Functions description:
*    Retrieves the file handle info records from the target
*/
function _UpdateTargetInfoFileHandles(NumVolumes) {
  var NextFileHandleAddr;
  var CurrentFileHandleInfoRecord;
  var Expression;
  var CurrentFileHandle;
  var CurrentFileHandleBufferPtr;
  var CurrentFileBuffer;
  var spNext;

  if (_Feature_EmFilePresent == true) {
    if (_Cache_FileHandleInfos_IsValid == false) {
      _Cache_FileHandleInfos = [];
      _UpdateTargetInfoFsGlobal();
      NextFileHandleAddr = _Cache_FS_Global["pFirstFileHandle"];
      do {
        CurrentFileHandleInfoRecord                      = {};
        CurrentFileHandleInfoRecord["FileHandle"]        = {};
        CurrentFileHandleInfoRecord["FileBufferPresent"] = false;
        CurrentFileHandleInfoRecord["FileBuffer"]        = {};
        CurrentFileHandleInfoRecord["Addr"]              = 0x0;
        if (NextFileHandleAddr != 0) {
          CurrentFileHandleInfoRecord["Addr"]       = NextFileHandleAddr;
          Expression                                = "*(FS_FILE*)" + _ToStrHex(NextFileHandleAddr);
          CurrentFileHandle                         = Debug.evaluate(Expression);
          CurrentFileHandleInfoRecord["FileHandle"] = CurrentFileHandle;
          CurrentFileHandleBufferPtr = CurrentFileHandle["pBuffer"];
          if (CurrentFileHandleBufferPtr == 0) {
            CurrentFileHandleInfoRecord["FileBufferPresent"] = false;
          } else {
            CurrentFileHandleInfoRecord["FileBufferPresent"] = true;
            Expression                 = "*(FS_FILE_BUFFER*)" + _ToStrHex(CurrentFileHandleBufferPtr);
            CurrentFileBuffer          = Debug.evaluate(Expression);
            CurrentFileHandleInfoRecord["FileBuffer"] = CurrentFileBuffer;
          }
          _Cache_FileHandleInfos.push(CurrentFileHandleInfoRecord);
          NextFileHandleAddr = CurrentFileHandle["pNext"];
        }
      } while (NextFileHandleAddr != 0);
      _Cache_FileHandleInfos_IsValid = true;
    }
  }
}

/*********************************************************************
*
* Page File Handles (horizontal)
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageFileHandlesHColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageFileHandlesHColHeader() {
  var Result;
  var NumFileHandles;

  Result = [];
  if (_Feature_EmFilePresent == true) {
    Result = ["Id", "In Use", "File Object", "Access Mode", "Cursor Pos", "Last Error", "File Buffer"];
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageFileHandlesHRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPageFileHandlesHRow(RowIdx) {
  var Result;
  var NumFileHandles;
  var FileHandleInfoStruct;
  var FileHandleInfo;
  var FileBufferInfo;
  var Value;
  var sValue;

  Result = [];
  if (RowIdx == 0) {
    _UpdateFeatureFlags();
  }
  if (_Feature_EmFilePresent == true) {
    if (RowIdx == 0) {
      _UpdateTargetInfoFileHandles();
      _UpdateTargetInfoFileObjects();
    }
    NumFileHandles = _Cache_FileHandleInfos.length;
    if (NumFileHandles > RowIdx) {
      FileHandleInfoStruct = _Cache_FileHandleInfos[RowIdx];
      FileHandleInfo       = FileHandleInfoStruct["FileHandle"];
      if (FileHandleInfoStruct["FileBufferPresent"] == true) {
        FileBufferInfo  = FileHandleInfoStruct["FileBuffer"];
      }
      //
      // ID (Address)
      //
      sValue               = _ToStrHex(FileHandleInfoStruct["Addr"]);
      Result.push (sValue);
      //
      // In Use
      //
      sValue               = _ToStrBoolYesNo(FileHandleInfo["InUse"]);
      Result.push (sValue);
      //
      // File Object (file name or address)
      //
      sValue               = _GetFileNameFromFileObjectAddr(FileHandleInfo["pFileObj"]);
      Result.push (sValue);
      //
      // Access Mode
      //
      sValue               = _ToStrFileHandleAccessMode(FileHandleInfo["AccessFlags"]);
      Result.push (sValue);
      //
      // Cursor Pos
      //
      sValue               = _ToStr(FileHandleInfo["FilePos"]);
      Result.push (sValue);
      //
      // Last Error
      //
      sValue               = _ToStr(FileHandleInfo["Error"]);
      Result.push (sValue);
      //
      // File Buffer
      //
      if(_Feature_FS_SUPPORT_FILE_BUFFER == true) {
        if (FileHandleInfoStruct["FileBufferPresent"] == true) {
          FileBufferInfo  = FileHandleInfoStruct["FileBuffer"];
          sValue          = "Position: " + FileBufferInfo["FilePos"];
          sValue         += ", size: "   + _ToStrByteSuffix(FileBufferInfo["BufferSize"]);
          sValue         += ", used: "   + _ToStrByteSuffix(FileBufferInfo["NumBytesInBuffer"]);
          sValue         += ", dirty: "  + _ToStrBoolYesNo(FileBufferInfo["IsDirty"]);
          sValue         += ", flags: "  + _ToStrFileBufferFlags(FileBufferInfo["Flags"]);
        } else {
          sValue = "<no buffer>";
        }
      } else {
        sValue = _GetFeatureNotEnabledStr();
      }
      Result.push (sValue);
    }
  }
  return Result;
}

/*********************************************************************
*
* Page File Objects
*
*********************************************************************/

/*********************************************************************
*
*       _UpdateTargetInfoFileHandles()
*
*  Functions description:
*    Retrieves the file handle info records from the target
*/
function _UpdateTargetInfoFileObjects(NumVolumes) {
  var NextFileObjectAddr;
  var CurrentFileObjectInfoRecord;
  var Expression;
  var CurrentFileObject;
  var CurrentCryptObjectAddr;
  var CurrentCryptObject;
  var CurrentFileObjectInfoRecord;
  
  if (_Feature_EmFilePresent == true) {
    if (_Cache_FileObjectInfos_IsValid == false) {
      _Cache_FileObjectInfos   = [];
      _UpdateTargetInfoFsGlobal();
      NextFileObjectAddr = _Cache_FS_Global["pFirstFileObj"];
      do {
        CurrentFileObjectInfoRecord                = {};
        CurrentFileObjectInfoRecord["FileObject"]  = {};
        CurrentFileObjectInfoRecord["Addr"]        = 0x0;
        CurrentFileObjectInfoRecord["HasCryptObj"] = false;
        CurrentFileObjectInfoRecord["CryptObj"]    = {};
        CurrentFileObjectInfoRecord["HasAlgoType"] = false;
        CurrentFileObjectInfoRecord["AlgoType"]    = {};
    
        if (NextFileObjectAddr != 0) {
          CurrentFileObjectInfoRecord["Addr"]       = NextFileObjectAddr;
          Expression                                = "*(FS_FILE_OBJ*)" + _ToStrHex(NextFileObjectAddr);
          CurrentFileObject                         = Debug.evaluate(Expression);
          CurrentFileObjectInfoRecord["FileObject"] = CurrentFileObject;
          if (_Feature_FS_SUPPORT_ENCRYPTION == true) {
            CurrentCryptObjectAddr                  = CurrentFileObject["pCryptObj"];
          } else {
            CurrentCryptObjectAddr                  = 0;
          }
          if (CurrentCryptObjectAddr != 0) {
            Expression                                 = "*(FS_CRYPT_OBJ*)" + _ToStrHex(CurrentCryptObjectAddr);
            CurrentCryptObject                         = Debug.evaluate(Expression);
            CurrentFileObjectInfoRecord["HasCryptObj"] = true;
            CurrentFileObjectInfoRecord["CryptObj"]    = CurrentCryptObject;
    
            CurrentAlgoTypeAddr                          = CurrentCryptObject["pAlgoType"];
            if (CurrentAlgoTypeAddr != 0) {
              Expression                                 = "*(FS_CRYPT_ALGO_TYPE*)" + _ToStrHex(CurrentAlgoTypeAddr);
              CurrentAlgoType                            = Debug.evaluate(Expression);
              CurrentFileObjectInfoRecord["HasAlgoType"] = true;
              CurrentFileObjectInfoRecord["AlgoType"]    = CurrentAlgoType;
            }
          }
          _Cache_FileObjectInfos.push(CurrentFileObjectInfoRecord);
          NextFileObjectAddr = CurrentFileObject["pNext"];
        }
      } while (NextFileObjectAddr != 0);
      _Cache_FileObjectInfos_IsValid = true;
    }
  }
}

/*********************************************************************
*
* Page File Objects (horizontal)
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageFileObjectHColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageFileObjectHColHeader() {
  var Result;
  var NumFileHandles;

  Result = [];
  _UpdateFeatureFlags();
  if (_Feature_EmFilePresent == true) {
    Result = ["Id", "Status", "File Name", "Size", "Volume", "Partition", "Encryption"];
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageFileObjectsHRow()
*
*  Functions description:
*    Delivers the requested data into the column of the respective page.
*/
function _GetPageFileObjectsHRow(RowIdx) {
  Result = []; 
  if(RowIdx == 0) {
    _UpdateFeatureFlags();
  }
  if(_Feature_EmFilePresent == true) {
    if(RowIdx == 0) {
      _UpdateTargetInfoVolume();      // Update volume information for volume name
      _UpdateTargetInfoFileObjects(); // Update file object information
    }
    NumFileObjects = _Cache_FileObjectInfos.length;
    if(NumFileObjects > RowIdx) {
      FileObjectInfoRecord = _Cache_FileObjectInfos[RowIdx];
      CurrentFileObject    = FileObjectInfoRecord["FileObject"];
      DirEntryPosInfo      = CurrentFileObject["DirEntryPos"];
      FatInfo              = DirEntryPosInfo["fat"];
      EfsInfo              = DirEntryPosInfo["efs"];
      //
      // Id
      //
      Result.push(_ToStrHex(FileObjectInfoRecord["Addr"]));
      if(CurrentFileObject["UseCnt"] > 0) {
        //
        // File object is used: Display details
        //
        sValue = _ToStrOrEmptyTag(CurrentFileObject["UseCnt"]);
        Result.push("Used by " + sValue + " handles");
        //
        //File Name
        //
        if(_Feature_FS_MULTI_HANDLE_SAFE == true) {
          sValue = _GetFileNameFromFileObject(CurrentFileObject);
          if(sValue == "") {
            sValue = "<No File Name>";
          }
        } else {
          sValue = _GetFeatureNotEnabledStr();
        }
        Result.push(sValue);
        //
        // Size
        //
        sValue = _ToStrByteSuffix(CurrentFileObject["Size"]);
        Result.push(sValue);
        //
        // Volume Information
        //
        sValue = _GetVolumeName(CurrentFileObject["pVolume"]);
        if (sValue == undefined) {
          sValue = "<invalid volume info>";
        }
        Result.push(sValue);
        //
        // Partition
        //
        if ((FatInfo["SectorIndex"]   != 0) || 
            (FatInfo["DirEntryIndex"] != 0)) {
          sValue = "FAT, Sector: " + FatInfo["SectorIndex"] + ", Position: " + FatInfo["DirEntryIndex"];
        } else if ((EfsInfo["FirstClusterId"] != 0) ||
                   (EfsInfo["DirEntryPos"]    != 0)) {
          sValue = "EFS, Sector: " + EfsInfo["FirstClusterId"] + ", Position: " + EfsInfo["DirEntryPos"];
        } else {
          sValue = "<No Partition Info>";
        }
        Result.push(sValue);
        //
        // Encryption
        //
        if(_Feature_FS_SUPPORT_ENCRYPTION == true) {
          if(FileObjectInfoRecord["HasCryptObj"] == true) {
            sValue              = "Encryption size: " + _ToStrByteSuffix(CurrentFileObject["SizeEncrypted"]);
            CurrentCryptObject  = FileObjectInfoRecord["CryptObj"];
            CryptContextAdr     = CurrentCryptObject["pContext"];
            sValue             += ", Context: " + Debug.getSymbol(CryptContextAdr);
            sValue             += ", Bytes per Block: " + _ToStrPowerOf2ByteSuffix(CurrentCryptObject["ldBytesPerBlock"]);
            CryptoAlgoTypeAddr  = CurrentCryptObject["pAlgoType"];
            sValue             += ", Crypto algorithm: " + Debug.getSymbol(CryptoAlgoTypeAddr);
          } else {
            sValue = "No encryption";
          }
        } else {
          sValue = _GetFeatureNotEnabledStr();
        }
        Result.push(sValue);
      } else {
        //
        // File object is not used: Omit details
        //
        Result.push("Not in use", "", "", "", "", "");
      }
      
    }
  }
  return Result;
}

/*********************************************************************
*
* Page Volume (common)
*
*********************************************************************/

/*********************************************************************
*
*       _ToStrMountType()
*
*  Functions description:
*    Translates a mount type into a string.
*/
function _ToStrMountType(MountType, AutoMount) {
  var Result;

  switch (MountType) {
    case 0:
      Result = "No";
      break;
    case 1:
      Result = "Yes (read-only)";
      break;
    case 2:
      Result = "Yes (write-only)";
      break;
    case 3:
      Result = "Yes (read/write)";
      break;
    default:
      Result = "Unknown (mount type " + MountType + ")";
      break;
  }

  if (AutoMount != 0) {
    Result += ", auto";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrWriteMode()
*
*  Functions description:
*    Translates a mount type into a string.
*/
function _ToStrWriteMode(WriteModeVolume, WriteModeFsGlobal) {
  var Result;

  switch (WriteModeVolume) {
    case "FS_WRITEMODE_SAFE":
      Result = "Safe";
      break;
    case "FS_WRITEMODE_MEDIUM":
      Result = "Medium";
      break;
    case "FS_WRITEMODE_FAST":
      Result = "Fast";
      break;
    case "FS_WRITEMODE_UNKNOWN":
      if (WriteModeFsGlobal != undefined) {
        Result = _ToStrWriteMode (WriteModeFsGlobal, undefined);
      } else {
        Result = "Unknown ( " + WriteModeVolume + ")";
      }
      break;
    default:
      Result = "Unknown ( " + WriteModeVolume + ")";
      break;
  }
  return Result;
}

/*********************************************************************
*
* Page Volume (horizontal)
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageVolumeVColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageVolumeHColHeader() {
  var Result;

  Result = ["Volume", "Address", "Partition index", "Mounted", "In use", "Write mode", "Free unused sectors", "File buffer flags", "Journal file name"];
  return Result;
}

/*********************************************************************
*
*       _GetPageVolumeHRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPageVolumeHRow(RowIdx) {
  var Result;
  var NumVolumes;
  var VolumeInfoRecord;
  var VolumeInfo;
  var Value;
  var sValue;

  Result = [];
  if (RowIdx == 0) {
    _UpdateFeatureFlags();
  }
  if (_Feature_EmFilePresent == true) {
    if (RowIdx == 0) {
      _UpdateTargetInfoFsGlobal();
      _UpdateTargetInfoVolume();
    }
    NumVolumes = _Cache_VolumeInfos.length;
    if (NumVolumes > RowIdx) {
      VolumeInfoRecord = _Cache_VolumeInfos[RowIdx];
      VolumeInfo       = VolumeInfoRecord["Volume"];
      //
      // Volume-ID
      //
      Value            = VolumeInfoRecord["Addr"];
      sValue           = _GetVolumeName(Value);
      Result.push (sValue);
      //
      // Address
      //
      Value            = VolumeInfoRecord["Addr"];
      sValue           = _ToStrHex(Value);
      Result.push (sValue);
      //
      // Partition Index
      //
      Value            = VolumeInfo["PartitionIndex"];
      sValue           = _ToStr(Value);
      Result.push (sValue);
      //
      // Mount type
      //
      sValue           = _ToStrMountType(VolumeInfo["MountType"], VolumeInfo["AllowAutoMount"]);
      Result.push (sValue);
      //
      // In use
      //
      sValue           = _ToStrBoolYesNo(VolumeInfo["InUse"]);
      Result.push (sValue);
      //
      // Write mode
      //
      sValue           = _ToStrWriteMode(VolumeInfo["WriteMode"], _Cache_FS_Global["WriteMode"]);
      Result.push (sValue);
      //
      // Free sector
      //
      if (_Feature_FS_SUPPORT_FREE_SECTOR == true) {
        sValue           = _ToStrBoolYesNo(VolumeInfo["FreeSector"]);
      } else {
        sValue           = _GetFeatureNotEnabledStr();
      }
      Result.push (sValue);
      //
      // File buffer flags
      //
      if (_Feature_FS_SUPPORT_FILE_BUFFER == true) {
        sValue           = _ToStrFileBufferFlags(VolumeInfo["FileBufferFlags"]);
      } else {
        sValue           = _GetFeatureNotEnabledStr();
      }
      Result.push (sValue);
      //
      // Journal file name
      //
      if (_Feature_FS_SUPPORT_JOURNAL_FILENAME == true) {
        sValue           = _ByteArrayToStrOrEmptyTag(VolumeInfo["acJournalFileName"]);
      } else {
        sValue           = _GetFeatureNotEnabledStr();
      }
      Result.push (sValue);
    }
    return Result;
  }
}

/*********************************************************************
*
* Page Partition (horizontal)
*
*********************************************************************/

/*********************************************************************
*
*       _GetPagePartitionHColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPagePartitionHColHeader () {
  var Result;

  Result = ["Volume", "Unit", "Initialized", "Sectors", "Device Type", "Journal", "Cache"];
  return Result;
}

/*********************************************************************
*
*       _GetPagePartitionHRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPagePartitionHRow(RowIdx) {
  var Result;
  var NumVolumes;
  var VolumeInfoRecord;
  var VolumeInfo;
  var PartitionInfo;
  var DeviceInfo;
  var DataInfo;
  var JournalInfo;
  var Value;
  var sValue;

  Result = [];
  if (RowIdx == 0) {
    _UpdateFeatureFlags();
  }
  if (_Feature_EmFilePresent == true) {
    if (RowIdx == 0) {
      _UpdateTargetInfoFsGlobal();
      _UpdateTargetInfoVolume();
    }
    NumVolumes = _Cache_VolumeInfos.length;
    if (NumVolumes > RowIdx) {
      VolumeInfoRecord = _Cache_VolumeInfos[RowIdx];
      VolumeInfo    = VolumeInfoRecord["Volume"];
      PartitionInfo = VolumeInfo["Partition"];
      DeviceInfo    = PartitionInfo["Device"];
      DataInfo      = DeviceInfo["Data"];
      if (_Feature_FS_SUPPORT_JOURNAL == true) {
        JournalInfo = DataInfo["JournalData"];
      }
      //
      // Volume-ID
      //
      Value            = VolumeInfoRecord["Addr"];
      sValue           = _GetVolumeName(Value);
      Result.push (sValue);
      //
      // Device Unit
      //
      Value      = DataInfo["Unit"];
      sValue     = _ToStr(Value);
      Result.push (sValue);
      //
      // Init State
      //
      Value  = DataInfo["IsInited"];
      sValue = _ToStrBoolYesNo(Value);
      Result.push (sValue);
      //
      // Sectors
      //
      Value         = PartitionInfo["NumSectors"];
      sValue        = _ToStrSectorSuffix(Value);
      Value         = PartitionInfo["StartSector"];
      sValue       += ", start sector: " + _ToStr(Value);
      Result.push (sValue);
      //
      // Device Type
      //
      Value      = DeviceInfo["pType"];
      sValue     = Debug.getSymbol(Value);
      Result.push (sValue);
      //
      // Journal
      //
      if (_Feature_FS_SUPPORT_JOURNAL == true) {
        Value       = JournalInfo["Unit"];
        sValue      = "Unit: " + Value;
        Value       = JournalInfo["IsActive"];
        sValue     += ", active: " + _ToStrBoolYesNo(Value);
        Value       = JournalInfo["IsInited"];
        sValue     += ", initialized: " + _ToStrBoolYesNo(Value);
        Value       = JournalInfo["IsNewDataLogged"];
        sValue     += ", new data: " + _ToStrBoolYesNo(Value);
        Value       = JournalInfo["IsTransactionNested"];
        sValue     += ", nested transaction: " + _ToStrBoolYesNo(Value);
      } else {
        sValue = _GetFeatureNotEnabledStr();
      }
      Result.push (sValue);
      //
      // Cache
      //
      if (_Feature_FS_SUPPORT_CACHE == true) {
        Value = DataInfo["pCacheAPI"];
        if (Value != 0) {
          sValue = "Present";
        } else {
          sValue = "Absent";
        }
      } else {
        sValue = _GetFeatureNotEnabledStr();
      }
      Result.push (sValue);
    }
    return Result;
  }
}

/*********************************************************************
*
* Page FAT Info (common)
*
*********************************************************************/

/*********************************************************************
*
*       _ToStrFatCluster()
*
*  Function description:
*    Translates value to string
*/
function _ToStrFatCluster(Value) {
  var Result;

  if(Value != 0xFFFFFFFF) {
    Result = _ToStrClusterSuffix(Value);
  } else {
    Result = "<empty>";
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrFatType()
*
*  Function description:
*    Translates value to string
*/
function _ToStrFatType(Value) {
  var Result;

  Result = "FAT";
  if(Value != 0) {
    Result += Value;
  }
  return Result;
}

/*********************************************************************
*
*       _ToStrFatSuffix()
*
*  Function description:
*    Translates value to string
*/
function _ToStrFatSuffix(Value) {
  var Result;

  Result = Value.toString();
  if (Value == 1) {
    Result += " FAT";
  } else {
    Result += " FATs";
  }
  return Result;
}

/*********************************************************************
*
* Page FS Info (Horizontal)
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageFsHColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageFsHColHeader () {
  var Result;

  Result = ["Volume", "Filesystem", "Size", "Allocation Table", "Root Directory", "1st Data"];
  return Result;
}

/*********************************************************************
*
*       _GetPageFsHRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPageFsHRow(RowIdx) {
  var Result;
  var NumVolumes;
  var VolumeInfoRecord;
  var VolumeInfo;
  var FatType;
  var Value;
  var sValue;

  Result = [];
  if (RowIdx == 0) {
    _UpdateFeatureFlags();
  }
  if (_Feature_EmFilePresent == true) {
    if (RowIdx == 0) {
      _UpdateTargetInfoFsGlobal();
      _UpdateTargetInfoVolume();
    }
    NumVolumes = _Cache_VolumeInfos.length;
    if (NumVolumes > RowIdx) {
      VolumeInfoRecord = _Cache_VolumeInfos[RowIdx];
      //
      // Volume-ID
      //
      Value  = VolumeInfoRecord["Addr"];
      sValue = _GetVolumeName(Value);
      Result.push (sValue);
      if (VolumeInfoRecord["IsFAT"] == true) {
        VolumeInfo       = VolumeInfoRecord["Volume"];
        FSInfo           = VolumeInfo["FSInfo"];
        FATInfo          = FSInfo["FATInfo"];
        //
        // Filesystem
        //
        Value   = FATInfo["FATType"];
        sValue  = _ToStrFatType(Value);
        Result.push (sValue);
        //
        // Size (# of sectors, bytes per sector, #of clusters, bytes per cluster)
        //
        Value   = FATInfo["NumSectors"] * FATInfo["BytesPerSector"];
        sValue  = _ToStrKMGByteSuffix(Value);
        sValue += " (";
        Value   = FATInfo["NumSectors"];
        sValue += _ToStrSectorSuffix(Value);
        Value   = FATInfo["BytesPerSector"];
        sValue += " @ " + _ToStrByteSuffix(Value);
        Value   = FATInfo["NumClusters"];
        sValue += " / " + _ToStrClusterSuffix(Value);
        Value   = FATInfo["BytesPerCluster"];
        sValue += " @ " + _ToStrByteSuffix(Value);
        sValue += ")";
        Result.push (sValue);
        //
        // Allocation Table (number, size, type)
        //
        Value   = FATInfo["NumFATs"];
        sValue  = _ToStrFatSuffix(Value);
        Value   = FATInfo["FATSize"];
        sValue += ", size: " + _ToStrSectorSuffix(Value);
        Result.push (sValue);
        //
        // Root Directory (Position, number of entries)
        //
        Value   = FATInfo["RootDirPos"];
        FatType = FATInfo["FATType"];
        sValue  = "Position: "
        if (FatType == 32) {
          sValue += _ToStrClusterPrefix(Value);
        } else {
          sValue += _ToStrSectorPrefix(Value);
        }
        if (FatType != 32) {
          //
          // FAT32 can extend the root directory, so an infinite number of entries
          // is possible here. Therefore omit the number of entries for FAT32.
          //
          Value   = FATInfo["RootEntCnt"];
          sValue += ", " + _ToStrEntrySuffix(Value);
        }
        Result.push (sValue);
        //
        // 1st Data
        //
        Value   = FATInfo["FirstDataSector"];
        sValue  = _ToStrSectorPrefix(Value);
        Result.push (sValue);
      } else {
        if (VolumeInfoRecord["IsEFS"] == true) {
          VolumeInfo       = VolumeInfoRecord["Volume"];
          FSInfo           = VolumeInfo["FSInfo"];
          EFSInfo          = FSInfo["EFSInfo"];
          //
          // Filesystem
          //
          Value   = FATInfo["FATType"];
          sValue  = "EFS";
          Result.push (sValue);
          //
          // Size (# of sectors, bytes per sector, #of clusters, bytes per cluster)
          //
          Value   = EFSInfo["NumSectors"] * EFSInfo["BytesPerSector"];
          sValue  = _ToStrKMGByteSuffix(Value);
          sValue += " (";
          Value   = EFSInfo["NumSectors"];
          sValue += _ToStrSectorSuffix(Value);
          Value   = EFSInfo["BytesPerSector"];
          sValue += " @ " + _ToStrByteSuffix(Value);
          Value   = EFSInfo["NumClusters"];
          sValue += " / " + _ToStrClusterSuffix(Value);
          Value   = EFSInfo["BytesPerCluster"];
          sValue += " @ " + _ToStrByteSuffix(Value);
          sValue += ")";
          Result.push (sValue);
          //
          // Allocation Table (size, type)
          //
          Value   = EFSInfo["EFSTableSize"];
          sValue  = "Size: " + _ToStrSectorSuffix(Value);
          Result.push (sValue);
          //
          // Root Directory (n/a)
          //
          sValue = _GetFeatureNotEnabledStr();
          Result.push (sValue);
          //
          // 1st Data
          //
          Value   = EFSInfo["FirstDataSector"];
          sValue  = _ToStrSectorPrefix(Value);
          Result.push (sValue);
        } else {
          Result.push("<Unknown file system>");
          Result.push("");
          Result.push("");
          Result.push("");
          Result.push("");
        }
      }
    }
  }
  return Result;
}

/*********************************************************************
*
* Page FS-API
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageFsApiNumColumns()
*
*  Functions description:
*    Delivers the number of columns of the respective page.
*/
function _GetPageFsApiNumColumns() {
  var Result;

  _UpdateTargetInfoVolume();
  Result = _Cache_VolumeInfos.length;
  if ((_Feature_FS_SUPPORT_MULTIPLE_FS == false) && 
      (Result                          == 0)) {
    //
    // feature disabled: Allow at least 1 column to display that fact
    //
    Result = 1;
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageFsApiColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageFsApiColHeader () {
  var Result;
  var NumVolumes;

  Result = ["Item"];
  _UpdateFeatureFlags();
  if (_Feature_EmFilePresent == true) {
    NumVolumes = _GetPageFsApiNumColumns();
    for (var Idx = 0; Idx < NumVolumes; Idx++) {
      Result.push("Volume " + (Idx + 1));
    }
}
  return Result;
}

/*********************************************************************
*
*       _GetPageFsApiNameRow()
*
*  Functions description:
*    Translates the value of the FS-API pointer into strings and
*    composes the row content.
*/
function _GetPageFsApiNameRow (ItemName) {
  var Result;
  var NumVolumes;
  var VolumeInfoRecord;
  var FsApiAddr;
  var sValue;

  Result = [];
  _UpdateFeatureFlags();
  if (_Feature_EmFilePresent == true) {
    Result.push (ItemName);
    NumVolumes = _GetPageFsApiNumColumns();
    if (_Feature_FS_SUPPORT_MULTIPLE_FS == true) {
      for (var Idx = 0; Idx < NumVolumes; Idx++) {
        VolumeInfoRecord = _Cache_VolumeInfos[Idx];
        if (VolumeInfoRecord["HasFsApi"] == true) {
          FsApiAddr = VolumeInfoRecord["FsApiAddr"];
          sValue    = Debug.getSymbol(FsApiAddr);
        } else {
          sValue    = "<empty>";
        }
        Result.push (sValue);
      }
      } else {
      Result = _GetFeatureNotEnabledVRow(ItemName, NumVolumes);
    }
  } else {
    Result.push (ItemName);
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageFsApiMemberRow()
*
*  Functions description:
*    Translates the values of a member of the respective data structure
*    into strings and composes the row content.
*/
function _GetPageFsApiMemberRow (ItemName, Member, ToStrFunction) {
  var Result;
  var NumVolumes;
  var VolumeInfoRecord;
  var FsApiInfo;
  var Value;
  var sValue;

  Result = [];
  _UpdateFeatureFlags();
  if (_Feature_EmFilePresent == true) {
    NumVolumes = _GetPageFsApiNumColumns();
    if (_Feature_FS_SUPPORT_MULTIPLE_FS == true) {
      Result.push(ItemName);
      for (var Idx = 0; Idx < NumVolumes; Idx++) {
        VolumeInfoRecord = _Cache_VolumeInfos[Idx];
        if (VolumeInfoRecord["HasFsApi"] == true) {
          FsApiInfo = VolumeInfoRecord["FsApi"];
          Value     = FsApiInfo[Member];
          sValue    = ToStrFunction(Value);
        } else {
          Result = "<empty>";
        }
        Result.push(sValue);
      }
    } else {
      Result = _GetFeatureNotEnabledVRow(ItemName, NumVolumes);
    }
  } else {
    Result.push(ItemName);
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageFsApiRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPageFsApiRow(RowIdx) {
  var Result;
  var NumVolumes;

  Result     = [];
  NumVolumes = _Cache_VolumeInfos.length;
  switch (RowIdx) {
    case 0:
      _UpdateTargetInfoVolume();

      Result = _GetPageFsApiNameRow("FS-API");
      break;
    case 1:
      Result = _GetPageFsApiMemberRow("Function CheckBootSector()", "pfCheckBootSector", _ToStrFuncNameOrUnknownTag);
      break;
    case 2:
      Result = _GetPageFsApiMemberRow("Function OpenFile()", "pfOpenFile", _ToStrFuncNameOrUnknownTag);
      break;
    case 3:
      Result = _GetPageFsApiMemberRow("Function CloseFile()", "pfCloseFile", _ToStrFuncNameOrUnknownTag);
      break;
    case 4:
      Result = _GetPageFsApiMemberRow("Function Read()", "pfRead", _ToStrFuncNameOrUnknownTag);
      break;
    case 5:
      Result = _GetPageFsApiMemberRow("Function Write()", "pfWrite", _ToStrFuncNameOrUnknownTag);
      break;
    case 6:
      Result = _GetPageFsApiMemberRow("Function Format()", "pfFormat", _ToStrFuncNameOrUnknownTag);
      break;
    case 7:
      Result = _GetPageFsApiMemberRow("Function OpenDir()", "pfOpenDir", _ToStrFuncNameOrUnknownTag);
      break;
    case 8:
      Result = _GetPageFsApiMemberRow("Function CloseDir()", "pfCloseDir", _ToStrFuncNameOrUnknownTag);
      break;
    case 9:
      Result = _GetPageFsApiMemberRow("Function ReadDir()", "pfReadDir", _ToStrFuncNameOrUnknownTag);
      break;
    case 10:
      Result = _GetPageFsApiMemberRow("Function RemoveDir()", "pfRemoveDir", _ToStrFuncNameOrUnknownTag);
      break;
    case 11:
      Result = _GetPageFsApiMemberRow("Function CreateDir()", "pfCreateDir", _ToStrFuncNameOrUnknownTag);
      break;
    case 12:
      Result = _GetPageFsApiMemberRow("Function DeleteDir()", "pfDeleteDir", _ToStrFuncNameOrUnknownTag);
      break;
    case 13:
      Result = _GetPageFsApiMemberRow("Function Rename()", "pfRename", _ToStrFuncNameOrUnknownTag);
      break;
    case 14:
      Result = _GetPageFsApiMemberRow("Function Move()", "pfMove", _ToStrFuncNameOrUnknownTag);
      break;
    case 15:
      Result = _GetPageFsApiMemberRow("Function SetDirEntryInfo()", "pfSetDirEntryInfo", _ToStrFuncNameOrUnknownTag);
      break;
    case 16:
      Result = _GetPageFsApiMemberRow("Function GetDirEntryInfo()", "pfGetDirEntryInfo", _ToStrFuncNameOrUnknownTag);
      break;
    case 17:
      Result = _GetPageFsApiMemberRow("Function SetEndOfFile()", "pfSetEndOfFile", _ToStrFuncNameOrUnknownTag);
      break;
    case 18:
      Result = _GetPageFsApiMemberRow("Function Unmount()", "pfUnmount", _ToStrFuncNameOrUnknownTag);
      break;
    case 19:
      Result = _GetPageFsApiMemberRow("Function GetDiskInfo()", "pfGetDiskInfo", _ToStrFuncNameOrUnknownTag);
      break;
    case 20:
      Result = _GetPageFsApiMemberRow("Function GetVolumeLabel()", "pfGetVolumeLabel", _ToStrFuncNameOrUnknownTag);
      break;
    case 21:
      Result = _GetPageFsApiMemberRow("Function SetVolumeLabel()", "pfSetVolumeLabel", _ToStrFuncNameOrUnknownTag);
      break;
    case 22:
      Result = _GetPageFsApiMemberRow("Function CreateJournalFile()", "pfCreateJournalFile", _ToStrFuncNameOrUnknownTag);
      break;
    case 23:
      Result = _GetPageFsApiMemberRow("Function OpenJournalFile()", "pfOpenJournalFile", _ToStrFuncNameOrUnknownTag);
      break;
    case 24:
      Result = _GetPageFsApiMemberRow("Function GetIndexOfLastSector()", "pfGetIndexOfLastSector", _ToStrFuncNameOrUnknownTag);
      break;
    case 25:
      Result = _GetPageFsApiMemberRow("Function CheckVolume()", "pfCheckVolume", _ToStrFuncNameOrUnknownTag);
      break;
    case 26:
      Result = _GetPageFsApiMemberRow("Function UpdateFile()", "pfUpdateFile", _ToStrFuncNameOrUnknownTag);
      break;
    case 27:
      Result = _GetPageFsApiMemberRow("Function SetFileSize()", "pfSetFileSize", _ToStrFuncNameOrUnknownTag);
      break;
    case 28:
      Result = _GetPageFsApiMemberRow("Function FreeSectors()", "pfFreeSectors", _ToStrFuncNameOrUnknownTag);
      break;
    case 29:
      Result = _GetPageFsApiMemberRow("Function GetFreeSpace()", "pfGetFreeSpace", _ToStrFuncNameOrUnknownTag);
      break;
    case 30:
      Result = _GetPageFsApiMemberRow("Function GetATInfo()", "pfGetATInfo", _ToStrFuncNameOrUnknownTag);
      break;
    case 31:
      Result = _GetPageFsApiMemberRow("Function CheckDir()", "pfCheckDir", _ToStrFuncNameOrUnknownTag);
      break;
    case 32:
      Result = _GetPageFsApiMemberRow("Function CheckAT()", "pfCheckAT", _ToStrFuncNameOrUnknownTag);
      break;
    case 33:
      Result = _GetPageFsApiMemberRow("Function ReadATEntry()", "pfReadATEntry", _ToStrFuncNameOrUnknownTag);
      break;
    case 34:
      Result = _GetPageFsApiMemberRow("Function SetDirEntryInfoEx()", "pfSetDirEntryInfoEx", _ToStrFuncNameOrUnknownTag);
      break;
    case 35:
      Result = _GetPageFsApiMemberRow("Function pfGetDirEntryInfoEx()", "pfGetDirEntryInfoEx", _ToStrFuncNameOrUnknownTag);
      break;
    default:
      Result = null;
      break;
  }
  return Result;
}

/*********************************************************************
*
* Page FAT Config
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageFatConfigColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageFatConfigColHeader () {
  var Result;

  Result = ["Item", "State"];
  return Result;
}

/*********************************************************************
*
*       _GetPageFatConfigMemberRow()
*
*  Functions description:
*    Translates the values of a member of the respective data structure
*    into strings and composes the row content.
*/
function _GetPageFatConfigMemberRow(ItemName, Value, ToStrFunction) {
  var Result;
  var sValue;

  Result = [];
  Result.push(ItemName);
  sValue = ToStrFunction(Value);
  Result.push(sValue);
  return Result;
}

/*********************************************************************
*
*       _GetPageFatConfigRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPageFatConfigRow(RowIdx) {
  var Result;

  switch (RowIdx) {
    case 0:
      _UpdateFeatureFlags();

      Result = ["Long File name support" , "not yet detected"];
      break;
    case 1:
      Result = _GetPageEfsConfigMemberRow("FS_SUPPORT_FAT", _Feature_FS_SUPPORT_FAT, _ToStrBoolEnabledDisable);
      break;
    case 2:
      Result = _GetPageFatConfigMemberRow("FS_FAT_USE_FSINFO_SECTOR", _Feature_FS_FAT_USE_FSINFO_SECTOR, _ToStrBoolEnabledDisable);
      break;
    case 3:
      Result = _GetPageFatConfigMemberRow("FS_FAT_UPDATE_DIRTY_FLAG", _Feature_FS_FAT_UPDATE_DIRTY_FLAG, _ToStrBoolEnabledDisable);
      break;
    case 4:
      Result = _GetPageFatConfigMemberRow("FS_FAT_SUPPORT_FREE_CLUSTER_CACHE", _Feature_FS_FAT_SUPPORT_FREE_CLUSTER_CACHE, _ToStrBoolEnabledDisable);
      break;
    case 5:
      Result = _GetPageFatConfigMemberRow("FS_FAT_OPTIMIZE_DELETE", _Feature_FS_FAT_OPTIMIZE_DELETE, _ToStrBoolEnabledDisable);
      break;
    case 6:
      Result = _GetPageFatConfigMemberRow("FS_FAT_OPTIMIZE_LINEAR_ACCESS", _Feature_FS_FAT_OPTIMIZE_LINEAR_ACCESS, _ToStrBoolEnabledDisable);
      break;
    case 7:
      Result = _GetPageFatConfigMemberRow("FS_FAT_PERMIT_RO_FILE_MOVE", _Feature_FS_FAT_PERMIT_RO_FILE_MOVE, _ToStrBoolEnabledDisable);
      break;
    case 8:
      Result = _GetPageFatConfigMemberRow("FS_FAT_LFN_LOWER_CASE_SHORT_NAMES", _Feature_FS_FAT_LFN_LOWER_CASE_SHORT_NAMES, _ToStrBoolEnabledDisable);
      break;
    case 9:
      Result = _GetPageFatConfigMemberRow("FS_FAT_CHECK_UNUSED_DIR_ENTRIES", _Feature_FS_FAT_CHECK_UNUSED_DIR_ENTRIES, _ToStrBoolEnabledDisable);
      break;
    default:
      Result = null;
      break;
  }
  return Result;
}

/*********************************************************************
*
* Page EFS Config
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageEfsConfigColHeader()
*
*  Functions description:
*    Delivers the column headers of the respective page.
*/
function _GetPageEfsConfigColHeader() {
  var Result;

  Result = ["Item", "Value/State"];
  return Result;
}

/*********************************************************************
*
*       _GetPageEfsConfigMemberRow()
*
*  Functions description:
*    Translates the values of a member of the respective data structure
*    into strings and composes the row content.
*/
function _GetPageEfsConfigMemberRow(ItemName, Value, ToStrFunction) {
  var Result;
  var sValue;

  Result = [];
  Result.push(ItemName);
  sValue = ToStrFunction(Value);
  Result.push(sValue);
  return Result;
}

/*********************************************************************
*
*       _GetPageEfsConfigGlobalMemberRow()
*
*  Functions description:
*    Translates the values of a member of the respective data structure
*    into strings and composes the row content.
*/
function _GetPageEfsConfigGlobalMemberRow(ItemName, Member, ToStrFunction) {
  var Result;
  var Value;
  var sValue;

  Result = [];
  Result.push(ItemName);
  _UpdateFeatureFlags();
  if (_Feature_EmFilePresent == true) {
    Value  = _Cache_FS_Global[Member];
    sValue = ToStrFunction(Value);
    Result.push (sValue);
  }
  return Result;
}

/*********************************************************************
*
*       _GetPageEfsConfigRow()
*
*  Functions description:
*    Returns a row for the respective page.
*/
function _GetPageEfsConfigRow(RowIdx) {
  var Result;

  switch (RowIdx) {
    case 0:
      _UpdateFeatureFlags();

      if ((_Feature_FS_SUPPORT_EFS                  == true) && 
          (_Feature_FS_EFS_SUPPORT_DIRENTRY_BUFFERS == true)) {
        Result = _GetPageEfsConfigGlobalMemberRow("Number of EFS directory entry buffers", "NumDirEntryBuffers", _ToStr);
      } else {
        Result = _GetFeatureNotEnabledVRow("Number of EFS directory entry buffers", 1);
      }
      break;
    case 1:
      Result = _GetPageEfsConfigMemberRow("FS_SUPPORT_EFS", _Feature_FS_SUPPORT_EFS, _ToStrBoolEnabledDisable);
      break;
    case 2:
      Result = _GetPageEfsConfigMemberRow("FS_EFS_SUPPORT_STATUS_SECTOR", _Feature_FS_EFS_SUPPORT_STATUS_SECTOR, _ToStrBoolEnabledDisable);
      break;
    case 3:
      Result = _GetPageEfsConfigMemberRow("FS_EFS_SUPPORT_FREE_CLUSTER_CACHE", _Feature_FS_EFS_SUPPORT_FREE_CLUSTER_CACHE, _ToStrBoolEnabledDisable);
      break;
    case 4:
      Result = _GetPageEfsConfigMemberRow("FS_EFS_SUPPORT_DIRENTRY_BUFFERS", _Feature_FS_EFS_SUPPORT_DIRENTRY_BUFFERS, _ToStrBoolEnabledDisable);
      break;
    case 5:
      Result = _GetPageEfsConfigMemberRow("FS_EFS_OPTIMIZE_DELETE", _Feature_FS_EFS_OPTIMIZE_DELETE, _ToStrBoolEnabledDisable);
      break;
    case 6:
      Result = _GetPageEfsConfigMemberRow("FS_EFS_CASE_SENSITIVE", _Feature_FS_EFS_CASE_SENSITIVE, _ToStrBoolEnabledDisable);
      break;
    case 7:
      Result = _GetPageEfsConfigMemberRow("FS_EFS_OPTIMIZE_LINEAR_ACCESS", _Feature_FS_EFS_OPTIMIZE_LINEAR_ACCESS, _ToStrBoolEnabledDisable);
      break;
    default:
      Result = null;
      break;
  }
  return Result;
}

/*********************************************************************
*
* Common page handling
*
*********************************************************************/

/*********************************************************************
*
*       _GetPageIdx()
*
*  Functions description:
*    Returns the Index of a given page.
*/
function _GetPageIdx(PageName) {
  var Result;
  var Pages;
  
  Pages  = getPages();
  Result = null;
  for (var Idx = 0; Idx < Pages.length; Idx++) {
    if (Pages[Idx] == PageName) {
      Result = Idx;
      break;
    }
  }
  return Result;
}

/*********************************************************************
*
*       _getCurrentRow()
*
*  Functions description:
*    Returns the values of the currently selected row for a given page.
*/
function _getCurrentRow(PageName) {
  var Result;
  var PageIdx;

  PageIdx = _GetPageIdx (PageName)
  switch (PageIdx) {
    case 0:
      Result = _GetPageGlobalRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 1:
      Result = _GetPageMemManagerRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 2:
      Result = _GetPageFileHandlesHRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 3:
      Result = _GetPageFileObjectsHRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 4:
      Result = _GetPageVolumeHRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 5:
      Result = _GetPagePartitionHRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 6:
      Result = _GetPageFsHRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 7:
      Result = _GetPageFsApiRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 8:
      Result = _GetPageFatConfigRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 9:
      Result = _GetPageEfsConfigRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 10:
      Result = _GetPageFeaturesRow(_CurrentRowOnPage[PageIdx]);
      break;
    case 11:
      Result = _GetPageScriptInfoRow(_CurrentRowOnPage[PageIdx]);
      break;
    default:
      Result = null;
      break;
  }
  return Result;
}

/*********************************************************************
*
*       Exposed functions
*
**********************************************************************
*/

/*********************************************************************
*
*       init()
*
*  Functions description:
*    This function... Not yet sure what it does...
*/
function init() {
  _InvalidateCaches();
  for (var PageIdx = 0; PageIdx < _NumPages; PageIdx++) {
    _CurrentRowOnPage[PageIdx] = 0;
  }
}

/*********************************************************************
*
*       getName()
*
*  Functions description:
*    Returns the name of the middleware this script supplies support
*    for.
*/
function getName() {
  return "emFile";
}

/*********************************************************************
*
*       getPages()
*
*  Functions description:
*    Returns the names of the pages this script is capable to display.
*/
function getPages() {
  var PageNames = ["Global",
                   "Global" + _PagePathSeparator + "Memory Manager",
                   "Global" + _PagePathSeparator + "File Handles",
                   "Global" + _PagePathSeparator + "File Objects",
                   "Volume",
                   "Volume" + _PagePathSeparator + "Partition",
                   "Volume" + _PagePathSeparator + "FS",
                   "Volume" + _PagePathSeparator + "FS-API",
                   "Config" + _PagePathSeparator + "FAT-Config",
                   "Config" + _PagePathSeparator + "EFS-Config",
                   "Config" + _PagePathSeparator + "Features",
                   "Script Info"];
  return PageNames;
}

/*********************************************************************
*
*       getColHeaders()
*
*  Functions description:
*    Returns the names of the colums for a given page.
*/
function getColHeaders(PageName) {
  var Result;
  var PageIdx;

  _MiddlewareDetection();
  PageIdx = _GetPageIdx (PageName)
  switch (PageIdx) {
    case 0:
      Result = _GetPageGlobalColHeader();
      break;
    case 1: 
      Result = _GetPageMemManagerColHeader();
      break;
    case 2:
      Result = _GetPageFileHandlesHColHeader();
      break;
    case 3:
      Result = _GetPageFileObjectHColHeader();
      break;
    case 4:
      Result = _GetPageVolumeHColHeader();
      break;
    case 5:
      Result = _GetPagePartitionHColHeader();
      break;
    case 6:
      Result = _GetPageFsHColHeader();
      break;
    case 7:
      Result = _GetPageFsApiColHeader();
      break;
    case 8:
      Result = _GetPageFatConfigColHeader();
      break;
    case 9:
      Result = _GetPageEfsConfigColHeader();
      break;
    case 10:
      Result = _GetPageFeaturesColHeader();
      break;
    case 11:
      Result = _GetPageScriptInfoColHeader();
      break;
    default:
      Result = null;
      break;
  }
  return Result;
}

/*********************************************************************
*
*       getFirstRow()
*
*  Functions description:
*    Returns the values of the 1st row for a given page.
*/
function getFirstRow(PageName) {
  var Result;
  var PageIdx;

  _UpdateFeatureFlags();
  PageIdx = _GetPageIdx(PageName);
  if (PageIdx != null) {
    _CurrentRowOnPage[PageIdx] = 0;
    Result = _getCurrentRow(PageName);
  } else {
    Result = null;
  }
  return Result;
}

/*********************************************************************
*
*       getNextRow()
*
*  Functions description:
*    Returns the values of the next row for a given page.
*/
function getNextRow(PageName) {
  var Result;
  var PageIdx;

  PageIdx = _GetPageIdx(PageName);
  if (PageIdx != null) {
    _CurrentRowOnPage[PageIdx] += 1;
    Result = _getCurrentRow(PageName);
  } else {
    Result = null;
  }
  return Result;
}

/*********************************************************************
*
*       onTargetChanged()
*
*  Functions description:
*    Handles the event of target state having changed.
*/
function onTargetChanged() {
  _InvalidateCaches();
}

/****** End Of File *************************************************/
